module fkfilter_mod

  use fft_data_mod
  implicit none
  public :: cmpxlinterp, kfilter, build_fspace
  contains

    subroutine cmpxlinterp(c,a,b)
      complex, dimension(:), intent(inout):: c
      integer, intent(in)                 :: a, b

      integer                             :: i

      do i=1, b-a-1
        c(a:b)=(c(a)*real(b-a-i)+c(b)*real(i))/real(b-a)
     end do

   end subroutine cmpxlinterp
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
   subroutine kfilter(c,vmin,eps, dw,k,fft_1d_id)
     complex, dimension(:,:), intent(inout):: c
     real   , dimension(:)  , intent(in)   :: k
     real                   , intent(in)   :: vmin, dw, eps
     type(fft_type)         , intent(in)   :: fft_1d_id

     real    :: dk, ok, w, kmax, a, b
     integer :: iw, ik, nw, nk

     nw = size(c,1)
     nk = size(c,2)

     do iw=1,nw
        kmax=2*dw*(iw-1)/vmin
        call fft_1d_data(fft_1d_id,FFTW_FORWARD, nk, c(iw,:))
        do ik=1, nk
           a = abs(k(ik))-kmax
           if(a>0) c(iw,ik) = c(iw,ik)*exp(-eps*a**2) 
        end do
     end do
     
   end subroutine kfilter
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  function build_fspace(sampling_interval, N) result(omega)
    ! N must be even!!! It is the number of elements in vector to be fft-ed
    ! sampling_interval is d1, d2, d3,etc

    integer, intent(in) :: N
    real,    intent(in) :: sampling_interval
    real                :: dom
    integer             :: i
    real, dimension(N)  :: omega

    if(mod(N,2)==1) then
       write(0,*) "The number", N, "is not even!"
       call seperr("Error in buildfspace")
    end if

    dom=1./(sampling_interval*real(N))

    do i=1,N/2+1
       omega(i)=dom*real(i-1)
    end do

    do i=N/2+2,N
       omega(i)=dom*real(i-N/2-1)-omega(N/2+1)
    end do

  end function build_fspace
end module fkfilter_mod
