program CHAVO
  ! Check AVO
  ! Input: n1xn2xn3 cube
  ! Axis 1 should be angle, 2 and 3 midpoint and depth, in any order
  ! Output is: 
  ! want=ig : out=2xn2xn3 cube where axis 1 holds intercept, then gradient
  ! want=ia : out=n1xn2xn3 cube with the interpolated avo trends
  ! want=rs : out=n1xn2xn3 - the residuals
  ! want=vr : out=1xn2xn3 variance of the residuals after subtracting avo trend

  use sep,  only: sep_init, sep_close, sep_read, sep_write, to_history
  use ooio, only: axis, readaxis, readcharpar, writeaxis
  implicit none

  real   , allocatable, dimension (:) :: R, v, writevec
  integer                             :: j
  type(axis)                          :: x, y, t ! t = theta is angle, not time
  real, parameter                     :: deg2rad = 0.017453292 !=pi/180
  real                                :: a, b, c, d, I, G
  character(len=4)                    :: want

  call sep_init
  t = readaxis(1)
  if(t%o + t%d*(t%n-1)>30) call seperr('Angles outside Shuey range')
  x = readaxis(2)
  y = readaxis(3)
  want = readcharpar('want',(/'ig','ia','rs','vr'/),4)

  if(want=='ig') then
     call to_history('n1',2)
     allocate(writevec(2))
  elseif(want=='vr') then 
     call to_history('n1',1)
     allocate(writevec(1))
  end if
  call sep_close

  allocate(R(t%n), v(t%n))

  v = sin(deg2rad*(t%o + (/0:t%n-1/)*t%d))**2
  a = sum(v)
  c = sum(v*v)

  do j=1, x%n*y%n
     R = 0
     call sep_read(R)
     b = sum(R)
     d = sum(R*v)
     G = (a*b-t%n*d)/(a*a-t%n*c)
     I = (b-a*G)/t%n

     if(want=='ig') then
        writevec(1)=I
        writevec(2)=G
        call sep_write(writevec)
     elseif(want=='ia') then
        call sep_write(I + G*v)
     elseif(want=='rs') then
        call sep_write(R - I - G*v)
     elseif(want=='vr') then
        R = (R - I - G*v)**2
        writevec(1) = sum(R)/t%n
        call sep_write(writevec)
     end if

  end do
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  deallocate(R, v)
  if(want=='ig'.or.want=='vr') deallocate(writevec)
end program CHAVO
