module lloyd_mod 
  use quantile_mod
  use sep
  use wei_par
  implicit none
  real, private, save :: min_r_pct,min_s_pct,minme,maxme
  integer :: ndiv,niter_over
  contains
  subroutine init_lloyd()
    integer :: mpierr
    if (wpar%impi.eq.0) then
      call from_param("min_region_pct",min_r_pct,1.)
    end if
    min_r_pct=min_r_pct/100.
    if (wpar%nmpi >0) then
      call MPI_BCAST(min_r_pct,1,MPI_REAL,0,MPI_COMM_WORLD,mpierr)
    end if
    if (wpar%impi.eq.0) then
      call from_param("min_slow_dev",min_s_pct,1.)
    end if
    min_s_pct=min_s_pct/100.
    if (wpar%nmpi >0) then
      call MPI_BCAST(min_s_pct,1,MPI_REAL,0,MPI_COMM_WORLD,mpierr)
    end if
    if (wpar%impi.eq.0) then
      call from_param("niter_lloyd",niter_over,-1) 
    end if
    if (wpar%nmpi >0) then
      call MPI_BCAST(niter_over,1,MPI_INTEGER,0,MPI_COMM_WORLD,mpierr)
    end if
  end subroutine 
  integer function lloyd_go(array,near,nn,bound,verb)
    real :: array(:),near(:),bound(:)
    integer :: niter
    integer :: ib,i,nn
    real :: smin,smax,ds,dist_old,dist_near,distort_local(nn)
    real :: junk(nn)
    complex :: j(nn)
    logical :: verb
    integer :: block(size(array)),region_count(nn)
!initial guess
    minme=minval(array)
    maxme=maxval(array)
    min_r_pct=min_r_pct*size(array)
    junk=0
    do ib=1,nn 
      i=real(ib)*real(size(array))/real(nn+1)
      near(ib)=quantile(i,array)
    end do 
    if (niter_over.eq.-1) then
      niter=30
    else
      niter=niter_over
    end if 
    ndiv=nn
    do i=1,niter 
      junk=i
      call find_bound(near,bound)
      call partition_space(array,bound,block)
      call compute_center(array,block,near,region_count)
      dist_near=compute_distort(array,block,near,distort_local,region_count&
        &)
      if (i< niter/2.) then
        call change_regions(array,block,near,bound,distort_local&
          &,region_count)
      end if
    end do 
    min_r_pct=min_r_pct/size(array)
    lloyd_go=ndiv
  end function 
  subroutine change_regions(array,block,near,bound,distort_local&
    &,region_count)
    real :: array(:),near(:),bound(:),distort_local(:)
    integer :: block(:),i,region_count(:)
!Get rid of regions with low count
    call rid_small(array,block,near,bound,distort_local,region_count) 
!Get rid of regions with close centers
    call rid_close(array,block,near,bound,distort_local,region_count)
  end subroutine 
  subroutine rid_close(array,block,near,bound,distort_local,region_count&
    &)
    real :: array(:),near(:),bound(:),distort_local(:),maxd,val
    integer :: block(:),i,region_count(:),ireg
    maxd=1000.
    ireg=-1
    do i=1,ndiv-1 
      val=(near(i+1) -near(i))/near(i)
      if (val< min_s_pct .and. val < maxd) then
        ireg=i
        if (distort_local(i+1)< distort_local(i)) then
          ireg=i+1
        end if
        maxd=val
      end if
    end do 
    if (ireg.ne.-1) then
      call replace_region(ireg,array,block,near,bound,distort_local&
        &,region_count)
    end if
  end subroutine 
  subroutine delete_region(ireg_del,array,block,near,bound,distort_local&
    &,region_count)
    integer :: ireg_del
    real :: array(:),near(:),bound(:),distort_local(:),minr,maxr
    integer  :: ireg_split,ib(1)
    integer :: block(:),i,region_count(:)
    if (ireg_del.ne.ndiv) then
      near(ireg_del: ndiv-1)=near(ireg_del+1:) !shift down the center
      bound(ireg_del: ndiv-1)=bound(ireg_del+1:) 
      !shift down the region boundaries
    end if
  end subroutine 
  subroutine replace_region(ireg_del,array,block,near,bound,distort_local&
    &,region_count)
    integer :: ireg_del
    real :: array(:),near(:),bound(:),distort_local(:),minr,maxr
    integer  :: ireg_split,ib(1),iold_reg
    integer :: block(:),i,region_count(:)
    real :: pct_close,maxd
    maxd=0.
    ireg_split=-1
    do i=1,ndiv 
      pct_close=1000.
      if (i >1) then
        pct_close=(near(i)-near(i-1))/near(i)
      end if
      if (i<ndiv) then
        pct_close=max(pct_close,(near(i+1)-near(i))/near(i+1))
      end if
      if (pct_close > min_s_pct) then
        if (distort_local(i)> maxd .and. i.ne.ireg_del .and. region_count&
          &(i)> min_r_pct ) then
          maxd=distort_local(i)
          ireg_split=i
        end if
      end if
    end do 
!DON'T ADD A POINT IF THE MAX DEVIATION REGION'S DEVIATION
! IS REALLY  SMALL
    if (maxd <  min_s_pct**2/10.) then
      ireg_split=-1
    end if
!DELETE REGION IF ITS CLOSE
    pct_close=10000.
    call delete_region(ireg_del,array,block,near,bound,distort_local&
      &,region_count)
    if (ireg_split>0 ) then
      call add_region(ireg_del,ireg_split,array,block,near,bound&
        &,distort_local,region_count)
    else
      ndiv=   ndiv-1
    end if
  end subroutine 
  subroutine add_region(ireg_del,ireg_split,array,block,near,bound&
    &,distort_local,region_count)
    integer :: ireg_del
    real :: array(:),near(:),bound(:),distort_local(:),minr,maxr
    integer  :: ireg_split,ib(1)
    integer :: block(:),i,region_count(:)
    real :: dist_near
    if (ireg_split > ireg_del) then
      ireg_split= ireg_split-1
    end if
    near(ireg_split+2:ndiv)=near(ireg_split+1:ndiv-1)
    if (ireg_split.eq.1) then
      minr=minme
    else
      minr=bound(ireg_split-1)
    end if 
    if (ireg_split.eq.ndiv-1) then
      maxr=maxme
    else
      maxr=bound(ireg_split)
    end if 
    near(ireg_split)=2./3.*minr +1./3. * maxr
    near(ireg_split+1)=1./3.*minr +2./3. * maxr
    call find_bound(near,bound)
    call partition_space(array,bound,block)
    call compute_center(array,block,near,region_count)
    dist_near=compute_distort(array,block,near,distort_local,region_count&
      &)
  end subroutine 
  subroutine rid_small(array,block,near,bound,distort_local,region_count&
    &)
    real :: array(:),near(:),bound(:),distort_local(:)
    integer :: block(:),i,region_count(:)
    do i=1,ndiv 
      if (region_count(i) < min_r_pct) then
        call replace_region(i,array,block,near,bound,distort_local&
          &,region_count)
      end if
    end do
  end subroutine 
  subroutine compute_center(array,block,near,ncount)
    real :: array(:),near(:),temp1,temp2
    integer :: block(:),i,ia,ncount(:),if
    real :: low,high
    ncount=0
    near=0
    do i=1,size(array)
      near(block(i))=  near(block(i))+array(i)
      ncount(block(i))=  ncount(block(i))+1
    end do 
    where(ncount>0)
      near=near/ncount
end where
    do i=1,ndiv 
      if (ncount(i).eq.0) then
        if (i.eq.1) then
          if=2
          do while (near(if).eq.0.)
            if= if+1
          end do 
          near(1)=near(if)
        else if (i.eq.ndiv) then
          if=ndiv-1
          do while (near(if).eq.0.)
            if= if-1
          end do 
          near(ndiv)=near(if)
        end if
        low=near(i-1)
        if=i+1
        do while (if < ndiv .and. near(if).eq.0. )
          if= if+1
        end do 
        if (if.ne.ndiv) then
          high=near(if)
        else
          if (near(ndiv).ne.0.) then
            high=near(if)
          else
            high=low
          end if
        end if 
        near(i)=0.5*(high+low)
      end if
    end do
  end subroutine 
  subroutine partition_space(array,bound,block)
    real :: array(:)
    real :: bound(:)
    integer  :: block(:),i,ib
    do i=1,size(array)
      block(i)=1
      do while (array(i)>= bound(block(i)))
        block(i)=    block(i)+1
      end do
    end do
  end subroutine 
  real function compute_distort(array,block,near,distort_local,nn)
    real :: array(:),distort_local(:)
    integer :: block(:) ,nn(:)
    real :: near(:)
    integer :: i
    compute_distort=0.
    distort_local=0.
    do i=1,size(array)
      distort_local(block(i))=    distort_local(block(i))+(array(i)-near&
        &(block(i)))**2
    end do 
    compute_distort=sum(distort_local)/size(array)
    where(nn>1)
      distort_local=distort_local/nn
end where
  end function 
  subroutine find_bound(near,bound)
    real :: near(:),bound(:)
    integer :: i
    do i=1,ndiv-1
      bound(i)=near(i+1)+near(i)
    end do 
    bound=bound/2.
    bound(ndiv)=1000000.
  end subroutine 
end module 
