#include "3drvgraphics.h"
extern int sepxargc;
extern char **sepxargv;
FILE *fl;

void graph(void)
{
  InitGL ();
  glutDisplayFunc(Display);
  glutReshapeFunc(Reshape);
  glutKeyboardFunc(Keyboard);
  glutMainLoop();
}

void InitGL (void)
{
  x_length=xmax-xmin;
  y_length=ymax-ymin;
  z_length=zmax-zmin;
  view_radius=sqrt(pow(x_length,2)+pow(y_length,2)+pow(z_length,2));
  x_view=0.707*view_radius*cos(view_angle);
  y_view=      view_radius*sin(view_angle);
  z_view=0.707*view_radius*cos(view_angle);
  
  glutInit(&sepxargc, sepxargv);
  glutInitDisplayMode (GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
  glutInitWindowSize(500,500);
  glutInitWindowPosition(0.0,0.0);
  glutCreateWindow("3D rays");
  
  if(usecolor) glClearColor (0.0, 0.0, 0.0, 0.0);
  else glClearColor (1.0, 1.0, 1.0, 0.0);
  glShadeModel (GL_FLAT);
}

void Reshape (int w, int h)
{
  glViewport (0, 0, (GLsizei) w, (GLsizei) h);
  glMatrixMode (GL_PROJECTION);
  glLoadIdentity();
  gluPerspective (30.0,1.0,4*z_view,0.0);
  glMatrixMode (GL_MODELVIEW);
}

void Display(void)
{
  GLfloat color;

  glClear (GL_COLOR_BUFFER_BIT);
  glPushMatrix ();
  
  gluLookAt (zoom*x_view,zoom*y_view,zoom*z_view,
	     0.0, 0.0, 0.0,
	     0.0, 1.0, 0.0);
  
  /* ROTATE FROM HERE DOWN */
  glRotatef (yangle,0.0,1.0,0.0);
  
  /*mobile axes*/
  glBegin (GL_LINES);
  glColor3f (0.0, 1.0, 1.0);

  plotxyz(x_length,0       ,0       ); plotxyz(x_length,z_length,0       );  
  plotxyz(0       ,0       ,y_length); plotxyz(0       ,z_length,y_length);
  plotxyz(x_length,0       ,y_length); plotxyz(x_length,z_length,y_length);

  plotxyz(x_length,0       ,0       ); plotxyz(x_length,       0,y_length);
  plotxyz(x_length,z_length,0       ); plotxyz(x_length,z_length,y_length);
  plotxyz(0       ,z_length,0       ); plotxyz(0,       z_length,y_length);

  plotxyz(0       ,z_length,y_length); plotxyz(x_length,z_length,y_length);
  plotxyz(0       ,       0,y_length); plotxyz(x_length,0       ,y_length);
  plotxyz(0       ,z_length,0       ); plotxyz(x_length,z_length,0       );

  /* Oz */  glColor3f (1.0, 0.0, 0.0);
  plotxyz(0,0,0); plotxyz(0,z_length,0);

  /* Oy */  glColor3f (0.0, 1.0, 0.0);
  plotxyz(0,0,0); plotxyz(0,0,y_length);

  /* Ox */  glColor3f (0.0, 0.0, 1.0);
  plotxyz(0,0,0); plotxyz(x_length,0,0);
  
  glEnd();
  
  if(rays) {
    glBegin (GL_LINES);
    if (usecolor) glColor3f(1.0,0.0,0.0);
    else glColor3f(0.0,0.0,0.0);
    for(ia=0;ia<na;ia++)
      for(ir=0;ir<nr;ir++)
	for(it=0;it<nt-1;it++) {
	  if(rap[ia][ir][it+1].x!=ABSENT)
	    {
	      plot_vertex(rap[ia][ir][it]);
	      plot_vertex(rap[ia][ir][it+1]);
	    }}
    glEnd();
  }/* end rays */
  
  if(wavefronts) {
    glBegin (GL_LINES);
    if (usecolor) glColor3f(0.0,1.0,1.0);
    else glColor3f(0.0,0.0,0.0);
    for(ir=0;ir<nr;ir++)
      for(ia=0;ia<na-1;ia++) {
	color=(GLfloat)ia/(GLfloat)(na-1);
	if (usecolor) glColor3f(0.0,color,1-color);
	else glColor3f(0.0,0.0,0.0);
	if(rap[ia][ir][wft].v!=ABSENT) {
	  plot_vertex(rap[ia  ][ir][wft]);
	  plot_vertex(rap[ia+1][ir][wft]);
	}
      }
    for(ia=0;ia<na;ia++) {
      color=(GLfloat)ia/(GLfloat)(na-1);
      if (usecolor) glColor3f(0.0,color,1-color);
      else glColor3f(0.0,0.0,0.0);
      for(ir=0;ir<nr-1;ir++) {
	if(rap[ia][ir][wft].v!=ABSENT) {
	  plot_vertex(rap[ia][ir  ][wft]);
	  plot_vertex(rap[ia][ir+1][wft]);
	}}
      if(rap[ia][ir][wft].v!=ABSENT) {
	plot_vertex(rap[ia][nr-1][wft]);
	plot_vertex(rap[ia][0   ][wft]);
      }
    }
    glEnd();
  }/* end wavefronts */

  if(aperture) {
    glBegin (GL_LINES);
    if (usecolor) glColor3f(1.0,1.0,1.0);
    else glColor3f(0.0,0.0,0.0);
    for(ir=0;ir<nr-1;ir++){
      if(rap[apt][ir][wft].v!=ABSENT) {
	plot_vertex(rap[apt][ir  ][wft]);
	plot_vertex(rap[apt][ir+1][wft]);
      }
    }
    if(rap[apt][ir][wft].v!=ABSENT){
      plot_vertex(rap[apt][nr-1][wft]);
      plot_vertex(rap[apt][0   ][wft]);
    }
    glEnd();
  }/* end aperture */
  
  if(cone) {
    glBegin (GL_LINES);
    if (usecolor) glColor3f(1.0,0.0,0.0);
    else glColor3f(0.0,0.0,0.0);
    for(ir=0;ir<nr;ir++) {
      for(it=0;it<nt-1;it++) {
	if(rap[apt][ray][it+1].x!=ABSENT) {
	  plot_vertex(rap[apt][ir][it]);
	  plot_vertex(rap[apt][ir][it+1]);
	}
      }}
    glEnd();
  }/* end cone */
    
  if(oneray) {
    glBegin (GL_LINES);
    if (usecolor) glColor3f(1.0,1.0,0.0);
    else glColor3f(0.0,0.0,0.0);
    for(it=0;it<nt-1;it++) {
      if(rap[apt][ray][it+1].x!=ABSENT) {
	plot_vertex(rap[apt][ray][it]);
	plot_vertex(rap[apt][ray][it+1]);
      }
    }
    glEnd(); 
  }/*end oneray*/
  
  fprintf(stderr,"it=%d ii=%d ia=%d \n",wft,apt,ray);

  glPopMatrix ();
  glutSwapBuffers ();
}

void Keyboard( unsigned char key, int x, int y )
{
  switch (key) {
  case 27:  /* ESCAPE */
    exit(0);
    break;
  case 'y':
    yangle += 5.0;
    if (yangle > 360.) yangle = yangle - 360.;
    glutPostRedisplay (); 
    break;
  case 'Y':
    yangle -= 5.0;
    if (yangle < 0.) yangle = yangle + 360.;
    glutPostRedisplay (); 
    break;
  case 'u':
    view_angle+= PI/24;
    if (view_angle > PI/2) view_angle=PI/2;
    x_view=0.707*view_radius*cos(view_angle);
    y_view=      view_radius*sin(view_angle);
    z_view=0.707*view_radius*cos(view_angle);
    glutPostRedisplay ();
    break;
  case 'd':
    view_angle-= PI/24;
    if (view_angle < -PI/2) view_angle=-PI/2;
    x_view=0.707*view_radius*cos(view_angle);
    y_view=      view_radius*sin(view_angle);
    z_view=0.707*view_radius*cos(view_angle);
    glutPostRedisplay ();
    break;
  case 'i':
    zoom -= 0.05;
    if (zoom < 0.05) zoom = 0.05;
    glutPostRedisplay (); 
    break;
  case 'o':
    zoom += 0.1;
    if (zoom > 5.0) zoom = 5.0;
    glutPostRedisplay (); 
    break;
  case 'r':
    if(rays==YES) rays=NO;
    else rays=YES;
    glutPostRedisplay (); 
    break;
  case 'w':
    if(wavefronts==YES) wavefronts=NO;
    else wavefronts=YES;
    glutPostRedisplay ();
    break;
  case 'a':
    if(aperture==YES) aperture=NO;
    else aperture=YES;
    glutPostRedisplay ();
    break;
  case 'c':
    if(cone==YES) cone=NO;
    else cone=YES;
    glutPostRedisplay ();
    break;
  case 'q':
    if(oneray==YES) oneray=NO;
    else oneray=YES;
    glutPostRedisplay ();
    break;
  case ']':
    if(wft<nt-1) wft++;
    else wft=nt-1;
    glutPostRedisplay ();
    break;
  case '[':
    if(wft>0) wft--;
    else wft=0;
    glutPostRedisplay ();
    break;
  case '}':
    if(apt<na-1) apt++;
    else apt=na-1;
    glutPostRedisplay ();
    break;
  case '{':
    if(apt>0) apt--;
    else apt=0;
    glutPostRedisplay ();
    break;
  case ')':
    if(ray<nr-1) ray++;
    else ray=0;
    glutPostRedisplay ();
    break;
  case '(':
    if(ray>0) ray--;
    else ray=nr-1;
    glutPostRedisplay ();
    break;
  case '!':
    fl=fopen("wft.dat","w");
    for(ir=0;ir<nr;ir++)
      for(ia=0;ia<na;ia++) {
	fprintf(fl,"%d %d %.2f %.2f %.2f\n",ir,ia,
		rap[ia][ir][wft].x,
		rap[ia][ir][wft].y,
		rap[ia][ir][wft].z);
      }
    fclose(fl);
  default:
    break;
  }
}

void plot_vertex(point3D P)
{
  plotxyz(P.x-xmin,P.z-zmin,P.y-ymin); 
}

void plotxyz(float x, float z, float y)
{
  float xtr=x_length/2,ytr=y_length/2,ztr=z_length/2;
  glVertex3f (x-xtr,-(z-ztr),-(y-ytr));
}




