#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include "hwt_struct.h"
#include "hwt_global.h"
#include "hwt_vectop.h"
#include "hwt_intsub.h"

void measure_rays(rays* pRR)
{
  int it,ii,ia;
  float ddx,ddy,ddz,ddl;
  
  for(ii=0;ii<pRR->ni;ii++) {
    for(ia=0;ia<pRR->na;ia++) {
      for(it=1;it<pRR->nt;it++) {
	ddx=pRR->c0[ii][ia][it].x-pRR->c0[ii][ia][it-1].x;
	ddy=pRR->c0[ii][ia][it].y-pRR->c0[ii][ia][it-1].y;
	ddz=pRR->c0[ii][ia][it].z-pRR->c0[ii][ia][it-1].z;
	ddl=sqrt(ddx*ddx+ddy*ddy+ddz*ddz);
	pRR->c0[ii][ia][it].l=pRR->c0[ii][ia][it-1].l+ddl;
      }}}
}

void tesselate(rays *pRR, triplets *ptrp, tetrahedra *ptet)
{
  group_rays(pRR,ptrp);
  get_T4(pRR,ptrp,ptet);
}

void tesselate_init(rays *pRR, triplets *ptrp, tetrahedra *ptet)
{
  int iT4;

  ptrp->nR3=2*(pRR->ni-1)*(pRR->na);
  ptrp->R3=(triplet*) malloc(ptrp->nR3*sizeof(triplet));

  ptrp->rayA.point=(point3D*) malloc(pRR->nt*sizeof(point3D));
  ptrp->rayB.point=(point3D*) malloc(pRR->nt*sizeof(point3D));
  ptrp->rayC.point=(point3D*) malloc(pRR->nt*sizeof(point3D));
  
  ptet->nT4=3*(pRR->nt-2)+1;
  ptet->T4=(tetrahedron**) malloc(ptet->nT4*sizeof(tetrahedron*));
  for(iT4=0;iT4<ptet->nT4;iT4++)
    ptet->T4[iT4]=(tetrahedron*) malloc(ptrp->nR3*sizeof(tetrahedron));
  
  /***********
    T4[i][j]
       ^  ^   
      T4  R3
   ***********/
}

void tesselate_close(triplets *ptrp, tetrahedra *ptet)
{
  int iT4;
  
  free(ptrp->R3);
  free(ptrp->rayA.point);
  free(ptrp->rayB.point);
  free(ptrp->rayC.point);

  for(iT4=0;iT4<ptet->nT4;iT4++)
    free(ptet->T4[iT4]);
  free(ptet->T4);
}

void group_rays(rays *pRR, triplets *ptrp)
{
  int iR3=0; 
  int ii,ia;

  for(ii=0;ii<pRR->ni-1;ii++) {
    for(ia=0;ia<pRR->na;ia++) {
      /* the loop will output 2 triplets at a time */
      /*  x  x       -> incl
	  x  */
      ptrp->R3[iR3].A_i=ii  ;    ptrp->R3[iR3].A_a=ia  ;
      ptrp->R3[iR3].B_i=ii  ;    ptrp->R3[iR3].B_a=(ia+1==pRR->na)?0:ia+1;
      ptrp->R3[iR3].C_i=ii+1;	 ptrp->R3[iR3].C_a=ia  ;
      iR3++;
      
      /*      x     -> incl 
	      x  x  */	  
      ptrp->R3[iR3].A_i=ii  ;	 ptrp->R3[iR3].A_a=(ia+1==pRR->na)?0:ia+1; 
      ptrp->R3[iR3].B_i=ii+1;	 ptrp->R3[iR3].B_a=(ia+1==pRR->na)?0:ia+1; 
      ptrp->R3[iR3].C_i=ii+1;	 ptrp->R3[iR3].C_a=ia;
      iR3++;
    }}
}

void get_R3(rays *pRR,triplets *ptrp, int iR3)
{
  int A_i,A_a;
  int B_i,B_a;
  int C_i,C_a;
  int it;
  
  A_i=ptrp->R3[iR3].A_i;  A_a=ptrp->R3[iR3].A_a;  
  B_i=ptrp->R3[iR3].B_i;  B_a=ptrp->R3[iR3].B_a;  
  C_i=ptrp->R3[iR3].C_i;  C_a=ptrp->R3[iR3].C_a;

  for(it=0;it<pRR->nt;it++) {
    ptrp->rayA.point[it]=pRR->c0[A_i][A_a][it];
    ptrp->rayB.point[it]=pRR->c0[B_i][B_a][it];
    ptrp->rayC.point[it]=pRR->c0[C_i][C_a][it];
  }
}

void get_T4(rays *pRR, triplets *ptrp, tetrahedra *ptet)
{
  int iR3=0;
  int ia,ib,ic;	
  char last='a';
  
  for(iR3=0;iR3<ptrp->nR3;iR3++) {
    int iT4=0;
    get_R3(pRR,ptrp,iR3); 

    out_T4(ptrp,ptet,iT4,iR3,1,1,1,'a',0); iT4++;

    for(ia=1,ib=1,ic=1;ia<pRR->nt-1 || ib<pRR->nt-1 || ic<pRR->nt-1;) {
      ia++; last='a'; out_T4(ptrp,ptet,iT4,iR3,ia,ib,ic,last,ia-1); iT4++;
      ib++; last='b'; out_T4(ptrp,ptet,iT4,iR3,ia,ib,ic,last,ib-1); iT4++;
      ic++; last='c'; out_T4(ptrp,ptet,iT4,iR3,ia,ib,ic,last,ic-1); iT4++;
    }}
}

void out_T4(triplets *ptrp, tetrahedra *ptet,
	    int iT4,int iR3,int ia,int ib,int ic,char last,int i_last)  
{
  /* A */
  ptet->T4[iT4][iR3].A_i=ptrp->R3[iR3].A_i;
  ptet->T4[iT4][iR3].A_a=ptrp->R3[iR3].A_a;
  ptet->T4[iT4][iR3].ia=ia;
  /* B */
  ptet->T4[iT4][iR3].B_i=ptrp->R3[iR3].B_i;
  ptet->T4[iT4][iR3].B_a=ptrp->R3[iR3].B_a;
  ptet->T4[iT4][iR3].ib=ib;
  /* C */
  ptet->T4[iT4][iR3].C_i=ptrp->R3[iR3].C_i;
  ptet->T4[iT4][iR3].C_a=ptrp->R3[iR3].C_a;
  ptet->T4[iT4][iR3].ic=ic;
  /* D */
  if(last=='a') {
    ptet->T4[iT4][iR3].D_i=ptrp->R3[iR3].A_i;
    ptet->T4[iT4][iR3].D_a=ptrp->R3[iR3].A_a;
    ptet->T4[iT4][iR3].id=i_last;
  }
  else if(last=='b') {
    ptet->T4[iT4][iR3].D_i=ptrp->R3[iR3].B_i;
    ptet->T4[iT4][iR3].D_a=ptrp->R3[iR3].B_a;
    ptet->T4[iT4][iR3].id=i_last;
  }
  else { 	/* (last=='c') */
    ptet->T4[iT4][iR3].D_i=ptrp->R3[iR3].C_i;
    ptet->T4[iT4][iR3].D_a=ptrp->R3[iR3].C_a;
    ptet->T4[iT4][iR3].id=i_last;
  }
}

void get_times_fast(rays *pRR,triplets *ptrp,tetrahedra *ptet,
		    cube *pTT,cube *pLL)
{
  int iR3,iT4;
  point3D *pA,*pB,*pC,*pD;
  int ai,aa,ia,bi,ba,ib,ci,ca,ic,di,da,id;

  for(iT4=0;iT4<ptet->nT4;iT4++) {
    for(iR3=0;iR3<ptrp->nR3;iR3++) {
      /*      fprintf(stderr,"iR3=%d(%d) iT4=%d(%d) \n",iR3,nR3,iT4,nT4);*/
      ai=ptet->T4[iT4][iR3].A_i; 
      aa=ptet->T4[iT4][iR3].A_a;
      ia=ptet->T4[iT4][iR3].ia;

      bi=ptet->T4[iT4][iR3].B_i;
      ba=ptet->T4[iT4][iR3].B_a;
      ib=ptet->T4[iT4][iR3].ib;

      ci=ptet->T4[iT4][iR3].C_i;
      ca=ptet->T4[iT4][iR3].C_a;
      ic=ptet->T4[iT4][iR3].ic;

      di=ptet->T4[iT4][iR3].D_i;
      da=ptet->T4[iT4][iR3].D_a;
      id=ptet->T4[iT4][iR3].id;

      pA=&pRR->c0[ai][aa][ia];
      pB=&pRR->c0[bi][ba][ib];
      pC=&pRR->c0[ci][ca][ic];
      pD=&pRR->c0[di][da][id];

      pA->t=pRR->ot+ia*pRR->st;
      pB->t=pRR->ot+ib*pRR->st;
      pC->t=pRR->ot+ic*pRR->st;
      pD->t=pRR->ot+id*pRR->st;

      if(ingrid(pTT,pA,pB,pC,pD) && volum(pA,pB,pC,pD)>0 ) {
	/*      if(ingrid(pTT,pA,pB,pC,pD) ) {*/
	slic3(pTT,pLL,pA,pB,pC,pD);
	/*	punchit(pTT,pLL,pA,pB,pC,pD);*/
      }
    }}
}

void slic3(cube *pTT,cube *pLL,point3D* pA, point3D* pB, point3D* pC, point3D* pD)
{
  /* find a vertical slice through the tetrahedron defined by A,B,C,D. */
  /* it finds either one or two triangles */
  /* the calls slic2 which continues */
  point3D *pp[4],K1,K2,K3;
  int nAx,nBx,nCx,nDx,ix;
  float xx,eps=pTT->d2/100000;

  /* order the points A-B-C-D, so that A has the smallest inline coordinate (X) */
  pp[0]=pA; pp[1]=pB; pp[2]=pC; pp[3]=pD;
  sort(pp,4,'x');
  pA=pp[0]; pB=pp[1]; pC=pp[2]; pD=pp[3];

  nAx=integerup(pA->x,pTT->o2,pTT->d2);
  nBx=integerdn(pB->x,pTT->o2,pTT->d2);
  nCx=integerdn(pC->x,pTT->o2,pTT->d2);
  nDx=integerdn(pD->x,pTT->o2,pTT->d2);

  /* it the T4 is not in the grid, skip it */
  if(ingrid(pTT,pA,pB,pC,pD)) {
    /* special case - ABCD are in the same plane, i.e. have the same X */
    if (nAx==nBx && nAx==nCx && nAx==nDx) {
      float distbc;
      
      xx=pTT->o2+nAx*pTT->d2;

      pp[0]=pA; pp[1]=pB; pp[2]=pC; pp[3]=pD;
      sort(pp,4,'y');
      pA=pp[0]; pB=pp[1]; pC=pp[2]; pD=pp[3];
      
      distbc=sqrt((pB->y-pC->y)*(pB->y-pC->y)+(pB->z-pC->z)*(pB->z-pC->z));
      
      if (!distbc) { slic2(pTT,pLL,pA,pB,pC,nAx);	slic2(pTT,pLL,pB,pC,pD,nAx); }
      else {         slic2(pTT,pLL,pB,pA,pD,nAx);	slic2(pTT,pLL,pC,pA,pD,nAx); }
    }
    else {
      
      /* from A to B*/
      for(ix=MAX(nAx,0);ix<=MIN(nBx,pTT->n2-1);ix++) {
	xx=pTT->o2+ix*pTT->d2;
	
	/* other special cases - pathological cases */
	/* when the points have the same coordinates */
	if (ABS((float)(pA->x-pB->x))<eps && ABS((float)(pA->x-pC->x))>eps) {
	  K1=lineval(pB,pC,xx,'x');  K2=lineval(pA,pC,xx,'x'); 
	  K3=lineval(pA,pD,xx,'x');
	}
	else if (ABS((float)(pA->x-pB->x))<eps && ABS((float)(pA->x-pC->x))<eps) {
	  K1=lineval(pB,pD,xx,'x');  K2=lineval(pC,pD,xx,'x'); 
	  K3=lineval(pA,pD,xx,'x');
	}
	else { /* finally a decent case */
	  K1=lineval(pA,pB,xx,'x');  K2=lineval(pA,pC,xx,'x'); 
	  K3=lineval(pA,pD,xx,'x');
	}
	slic2(pTT,pLL,&K1,&K2,&K3,ix);
      }
      
      /* from B to C*/
      for(ix=MAX(nBx+1,0);ix<=MIN(nCx,pTT->n2-1);ix++) {
	xx=pTT->o2+ix*pTT->d2;

	K1=lineval(pA,pD,xx,'x');  K2=lineval(pB,pC,xx,'x');
	
	K3=lineval(pA,pC,xx,'x');  slic2(pTT,pLL,&K1,&K2,&K3,ix);
	K3=lineval(pB,pD,xx,'x');  slic2(pTT,pLL,&K1,&K2,&K3,ix);
      }
      
      /* from C to D*/
      for(ix=MAX(nCx+1,0);ix<=MIN(nDx,pTT->n2-1);ix++) {
	xx=pTT->o2+ix*pTT->d2;

	K1=lineval(pD,pA,xx,'x');  K2=lineval(pD,pB,xx,'x'); 
	K3=lineval(pD,pC,xx,'x');  slic2(pTT,pLL,&K1,&K2,&K3,ix);
      }

    } /* end if-else for the special case */
  } /* end ingrid */
}

void slic2(cube *pTT,cube *pLL,point3D* pA, point3D* pB, point3D* pC, int ix)
{
  point3D *pp[3];
  int nAy,nBy,nCy,iy;

  /* order K1,K2,K3 from slic3 by the xline coordinate */
  pp[0]=pA; pp[1]=pB; pp[2]=pC;
  sort(pp,3,'y');
  pA=pp[0]; pB=pp[1]; pC=pp[2];

  nAy=integerup(pA->y,pTT->o3,pTT->d3);
  nBy=integerdn(pB->y,pTT->o3,pTT->d3);
  nCy=integerdn(pC->y,pTT->o3,pTT->d3);

  for(iy=MAX(nAy,1);iy<=MIN(nBy,pTT->n3-1);iy++)
    slic1(pTT,pLL,pA,pB,pC,ix,iy);
  
  for(iy=MAX(nBy+1,1);iy<=MIN(nCy,pTT->n3-1);iy++)
    slic1(pTT,pLL,pC,pB,pA,ix,iy);
}

void slic1(cube *pTT,cube *pLL,
	   point3D* pA, point3D* pB, point3D* pC, int ix, int iy)
{
  point3D *pp[3];
  int izl,izh,iz;
  point3D M;
  float z1,z2,zl,zh,yy,zz;
  float tl,th,ti, ll,lh,li;

  int ggg;
  ggg=0;

  M.x=pTT->o2+ix*pTT->d2;
  M.y=pTT->o3+iy*pTT->d3; yy=M.y;

  if (pA->y==pB->y && pA->y!=pC->y) {
    z1=(lineval(pC,pB,yy,'y')).z;
    z2=(lineval(pC,pA,yy,'y')).z;
    /*    if(ix==40 && iy==0) wr(".1.");*/
  }
  else if (pA->y==pB->y && pA->y==pC->y) {
    pp[0]=pA; pp[1]=pB; pp[2]=pC;
    sort(pp,3,'z');
    pA=pp[0]; pB=pp[1]; pC=pp[2];
    z1=pA->z;
    z2=pC->z;
    /*    if(ix==40 && iy==0) wr(".2.");*/
  }
  else {
    z1=(lineval(pA,pB,yy,'y')).z;
    z2=(lineval(pA,pC,yy,'y')).z;
    /*    if(ix==40 && iy==0) wr(".3.");*/
  }

  izl=integerup(MIN(z1,z2),pTT->o1,pTT->d1);
  zl=pTT->o1+izl*pTT->d1;

  izh=integerdn(MAX(z1,z2),pTT->o1,pTT->d1);
  zh=pTT->o1+izh*pTT->d1;

  /*  if (izl<0) fprintf(stderr,"yy=%f\n",yy);*/
  /*  if (izl<0) fprintf(stderr,"xA=%f yA=%f zA=%f\n", pA->x, pA->y, pA->z);*/
  /*  if (izl<0) fprintf(stderr,"xB=%f yB=%f zB=%f\n", pB->x, pB->y, pB->z);*/
  /*  if (izl<0) fprintf(stderr,"xC=%f yC=%f zC=%f\n", pC->x, pC->y, pC->z);*/
  /*  if (izl<0) fprintf(stderr,"xM=%f yM=%f\n", M.x, M.y);*/
  /*  if (izl<0) fprintf(stderr,"L izl=%d ix=%d iy=%d\n",izl,ix,iy);*/
  /*  if (izl<0) fprintf(stderr,"L z1=%f z2=%f \n",z1,z2);*/

  if(izl<=izh) {
    /* smallest z - Zlow */
    M.z=zl;                    
    getit2(pTT,pA,pB,pC,&M); tl=M.t; ll=M.l;

    if (izl>=0 && izl<=pTT->n1) {
      if(pLL->c0[iy][ix][izl]==0.0) {
	pTT->c0[iy][ix][izl]=tl; 
	if(ggg) pTT->c0[iy][ix][izl]=1;
	pLL->c0[iy][ix][izl]=ll;
      }
      else
	if(ll<pLL->c0[iy][ix][izl]) {
	  pTT->c0[iy][ix][izl]=tl; 
	  if(ggg) pTT->c0[iy][ix][izl]=1;
	  pLL->c0[iy][ix][izl]=ll;
	}
    }
    
    /* biggest z - Zhigh */
    M.z=zh;                    
    getit2(pTT,pA,pB,pC,&M); th=M.t; lh=M.l;

    if(izh>=0 && izh<pTT->n1) {
      if(pLL->c0[iy][ix][izh]==0.0) {
	pTT->c0[iy][ix][izh]=th; 
	if(ggg) pTT->c0[iy][ix][izh]=10;
	pLL->c0[iy][ix][izh]=lh;
      }
      else
	if(lh<pLL->c0[iy][ix][izh]) {
	  pTT->c0[iy][ix][izh]=th; 
	  if(ggg) pTT->c0[iy][ix][izh]=10;
	  pLL->c0[iy][ix][izh]=lh;
	}
    }

    for(iz=MAX(izl+1,0);iz<=MIN(izh-1,pTT->n1-1);iz++) {
      zz=pTT->o1+iz*pTT->d1;
      li=ll+(lh-ll)*(zz-zl)/(zh-zl);
      ti=tl+(th-tl)*(zz-zl)/(zh-zl);
      if(pLL->c0[iy][ix][iz]==0.0) {
	pTT->c0[iy][ix][iz]=ti;
	if(ggg) pTT->c0[iy][ix][iz]=5;
	pLL->c0[iy][ix][iz]=li;
      }
      else
	if(li<pLL->c0[iy][ix][iz]) {
	  pTT->c0[iy][ix][iz]=ti;
	  if(ggg) pTT->c0[iy][ix][iz]=5;
	  pLL->c0[iy][ix][iz]=li;
	}
    }
  } /* end if izl<izh */
}

void getit2(cube *pTT,point3D *pA,point3D *pB,point3D *pC,point3D *pP)
{
  float ap[5],ab[5],ac[5];
  float detyz,detyt,detzt,detyl,detzl;
  float d, dmax;
  point3D *pp[3];
  point3D *pL,*pH;
  float eps=pTT->d2/100000;

  ab[1]=pB->y-pA->y; ac[1]=pC->y-pA->y; ap[1]=pP->y-pA->y; /*y*/
  ab[2]=pB->z-pA->z; ac[2]=pC->z-pA->z; ap[2]=pP->z-pA->z; /*z*/
  ab[3]=pB->t-pA->t; ac[3]=pC->t-pA->t;                    /*t*/
  ab[4]=pB->l-pA->l; ac[4]=pC->l-pA->l;                    /*l*/

  detyz=ab[1]*ac[2]-ab[2]*ac[1]; /*1-2 = y-z*/
  detyt=ab[1]*ac[3]-ab[3]*ac[1]; /*1-3 = y-t*/
  detzt=ab[2]*ac[3]-ab[3]*ac[2]; /*2-3 = z-t*/

  detyl=ab[1]*ac[4]-ab[4]*ac[1]; /*1-3 = y-l*/
  detzl=ab[2]*ac[4]-ab[4]*ac[2]; /*2-3 = z-l*/

  if(ABS(detyz)>eps) {
    pP->t=pA->t+ap[2]*detyt/detyz-ap[1]*detzt/detyz;
    pP->l=pA->l+ap[2]*detyl/detyz-ap[1]*detzl/detyz;
  }
  else {
    pp[0]=pA; pp[1]=pB; pp[2]=pC;
    sort(pp,3,'z');
    pL=pp[0]; pH=pp[2];

    dmax=sqrt((pH->z-pL->z)*(pH->z-pL->z)+(pH->y-pL->y)*(pH->y-pL->y));
    if (ABS(dmax)<eps) {
      pP->t=pL->t;
      pP->l=pL->l;
    }
    else {
      d=sqrt((pP->z-pL->z)*(pP->z-pL->z)+(pP->y-pL->y)*(pP->y-pL->y));
      pP->t=pL->t+(pH->t-pL->t)*d/dmax;
      pP->l=pL->l+(pH->l-pL->l)*d/dmax;
    }
  }
}

void sort(point3D** pp, int n, char c)
{
  int i,j,min;
  point3D *pt;

  for(i=0;i<n-1;i++) {
    min=i;
    for(j=i+1;j<n;j++)
      if (c=='x'){       if(pp[j]->x < pp[min]->x) min=j;}
      else if (c=='y') { if(pp[j]->y < pp[min]->y) min=j;}
      else {             if(pp[j]->z < pp[min]->z) min=j;}  
    pt=pp[min];pp[min]=pp[i];pp[i]=pt;
  }
}

point3D lineval(point3D* pA, point3D* pB, float vv,char c)
{
  point3D M;
  float ratio;

  if (c=='x') {
    ratio=(vv-pA->x)/(pB->x-pA->x);
    M.x=vv;
    M.y=pA->y+(pB->y-pA->y)*ratio;
  }
  else {
    ratio=(vv-pA->y)/(pB->y-pA->y);
    M.x=pA->x+(pB->x-pA->x)*ratio;
    M.y=vv;
  }
  M.z=pA->z+(pB->z-pA->z)*ratio;
  M.t=pA->t+(pB->t-pA->t)*ratio;
  M.l=pA->l+(pB->l-pA->l)*ratio;

  return M;
}

int ingrid(cube *pTT,point3D *pA,point3D *pB,point3D *pC,point3D *pD)
{ 
  int nAx,nAy,nAz,nBx,nBy,nBz,nCx,nCy,nCz,nDx,nDy,nDz;
  int xout,yout,zout;
  int mx,my,mz;

  mx=0;my=0;mz=0;

  nAx=integerup(pA->x,pTT->o2,pTT->d2);
  nBx=integerdn(pB->x,pTT->o2,pTT->d2);
  nCx=integerdn(pC->x,pTT->o2,pTT->d2);
  nDx=integerdn(pD->x,pTT->o2,pTT->d2);
  xout=(nAx>pTT->n2-1 && nBx>pTT->n2-1 &&  nCx>pTT->n2-1 && nDx>pTT->n2-1)|| 
    (   nAx<mx        && nBx<mx        &&  nCx<mx        && nDx<mx    );
  
  if (!xout) {
    nAy=integerup(pA->y,pTT->o3,pTT->d3);
    nBy=integerdn(pB->y,pTT->o3,pTT->d3);
    nCy=integerdn(pC->y,pTT->o3,pTT->d3);
    nDy=integerdn(pD->y,pTT->o3,pTT->d3);
    yout=(nAy>pTT->n3-1 && nBy>pTT->n3-1 &&  nCy>pTT->n3-1 && nDy>pTT->n3-1)|| 
      (   nAy<my        && nBy<my        &&  nCy<my        && nDy<my    );

    if (!yout) {
      nAz=integerup(pA->z,pTT->o1,pTT->d1);
      nBz=integerdn(pB->z,pTT->o1,pTT->d1);
      nCz=integerdn(pC->z,pTT->o1,pTT->d1);
      nDz=integerdn(pD->z,pTT->o1,pTT->d1);
      zout=(nAz>pTT->n1-1 && nBz>pTT->n1-1 && nCz>pTT->n1-1 && nDz>pTT->n1-1)||
	(   nAz<mz        && nBz<mz        &&  nCz<mz        && nDz<mz   );
    }
  }
  
  if( xout || yout || zout)
    return NO;
  else
    return YES;
}

void median_filter(cube *pTT, cube *pLL)
{
  int i1,i2,i3;
  int j1,j2,j3;
  float pm[27];
  float pmm;
  int i,j,k;
  int min;

  for(i3=0;i3<pTT->n3;i3++) {
    for(i2=0;i2<pTT->n2;i2++) {
      for(i1=0;i1<pTT->n1;i1++) {
	pLL->c0[i3][i2][i1]=pTT->c0[i3][i2][i1];
      }}}
  
  for(i3=0;i3<pTT->n3;i3++) {
    for(i2=0;i2<pTT->n2;i2++) {
      for(i1=0;i1<pTT->n1;i1++) {
	if (pTT->c0[i3][i2][i1]==NOTIME) {
	  k=0;
	  for(j3=MAX(i3-1,0);j3<=MIN(i3+1,pTT->n3-1);j3++) {
	    for(j2=MAX(i2-1,0);j2<=MIN(i2+1,pTT->n2-1);j2++) {
	      for(j1=MAX(i1-1,0);j1<=MIN(i1+1,pTT->n1-1);j1++) {
		if (pLL->c0[j3][j2][j1] != NOTIME) {
		  pm[k]=pLL->c0[j3][j2][j1];
		  k++;
		}
	      }}}
	  if (k!=0) {
	    /* sort by magnitude*/
	    for(i=0;i<k-1;i++) {
	      min=i;
	      for(j=i+1;j<k;j++) {
		if(pm[j] < pm[min]) min=j;
	      }
	      pmm=pm[min]; pm[min]=pm[i]; pm[i]=pmm;
	    }
	    pTT->c0[i3][i2][i1]=pm[k/2];
	  } /* end sort */
	} /* end if NOTIME */
    }}}
}

void mean_filter(cube *pTT, cube *pLL)
{
  int i1,i2,i3;
  int j1,j2,j3;
  int k;
  float sum;

  for(i3=0;i3<pTT->n3;i3++) {
    for(i2=0;i2<pTT->n2;i2++) {
      for(i1=0;i1<pTT->n1;i1++) {
	pLL->c0[i3][i2][i1]=pTT->c0[i3][i2][i1];
      }}}

  for(i3=0;i3<pTT->n3;i3++) {
    for(i2=0;i2<pTT->n2;i2++) {
      for(i1=0;i1<pTT->n1;i1++) {
	k=0;
	sum=0;
	for(j3=MAX(i3-1,0);j3<=MIN(i3+1,pTT->n3-1);j3++) {
	  for(j2=MAX(i2-1,0);j2<=MIN(i2+1,pTT->n2-1);j2++) {
	    for(j1=MAX(i1-1,0);j1<=MIN(i1+1,pTT->n1-1);j1++) {
	      sum+=pLL->c0[j3][j2][j1];
	      k++;
	    }}}
	pTT->c0[i3][i2][i1]=sum/k;
      }}}
}

/*void punchit(cube *pTT,cube *pLL,*/
/*	     point3D* pA, point3D* pB, point3D* pC, point3D* pD)*/
/*{*/
/*  float  min2,  max2,  min3,  max3;*/
/*  int   nmin2, nmax2, nmin3, nmax3;*/
/*  int i2,i3;*/
/*  point3D M1,M2;*/
/*  int in;*/

/*  getdiag(pA,pB,pC,pD);*/
/*  */
/*  min2=MIN(pD->x,MIN(pC->x,MIN(pB->x,pA->x))); nmin2=(min2-pTT->o2)/pTT->d2;*/
/*  max2=MAX(pD->x,MAX(pC->x,MAX(pB->x,pA->x))); nmax2=(max2-pTT->o2)/pTT->d2;*/
/**/
/*  min3=MIN(pD->y,MIN(pC->y,MIN(pB->y,pA->y))); nmin3=(min3-pTT->o3)/pTT->d3;*/
/*  max3=MAX(pD->y,MAX(pC->y,MAX(pB->y,pA->y))); nmax3=(max3-pTT->o3)/pTT->d3;*/
/**/
/*  for(i2=nmin2;i2<=nmax2;i2++) {*/
/*    M1.x=pTT->o2+i2*pTT->d2; M2.x=M1.x;*/
/*    for(i3=nmin3;i3<=nmax3;i3++) {*/
/*      wr(".");*/
/*      M1.y=pTT->o3+i3*pTT->d3; M2.y=M2.y;*/
/*      */
/*      in=NO;*/
/*      if(intri(&M1,pA,pB,pC)) {*/
/*	punch(&M1,pA,pB,pC);*/
/*	in=YES;*/
/*      }*/
/*      else if (intri(&M1,pA,pB,pD)) {*/
/*	punch(&M1,pA,pB,pD);*/
/*	in=YES;*/
/*      }*/
/*      else {*/
/*	;*/
/*      }*/
/*      */
/*      if (in) {*/
/*	if(intri(&M2,pC,pD,pA)) {*/
/*	  punch(&M2,pC,pD,pA);*/
/*	}*/
/*	else {*/
/*	  punch(&M2,pC,pD,pB);*/
/*	}*/
/*      }*/
/*    }}*/
/*wr("\n");*/
/**/
/*}*/

/* void punch(point3D* pT, point3D* pP, point3D* pQ, point3D* pR) */
/* { */
/*   float m[4]; */
/*   float xQP,yQP,zQP,tQP,lQP, xQR,yQR,zQR,tQR,lQR; */
/*   float dxy, dyz,dxz, dyt,dxt, dyl,dxl; */

/*   xQP=pQ->x-pP->x;  xQR=pQ->x-pR->x; */
/*   yQP=pQ->y-pP->y;  yQR=pQ->y-pR->y; */
/*   zQP=pQ->z-pP->z;  zQR=pQ->z-pR->z; */
/*   tQP=pQ->t-pP->t;  tQR=pQ->t-pR->t; */
/*   lQP=pQ->l-pP->l;  lQR=pQ->l-pR->l; */

/*   m[0]=xQP; m[1]=yQP; */
/*   m[2]=xQR; m[3]=yQR; dxy=det2(m); */
  
/*   m[0]=yQP; m[1]=zQP; */
/*   m[2]=yQR; m[3]=zQR; dyz=det2(m); */
/*   m[0]=xQP; m[1]=zQP; */
/*   m[2]=xQR; m[3]=zQR; dxz=det2(m); */
/*   pT->z=pQ->z - (pT->x-pQ->x)*dyz/dxy + (pT->y-pQ->y)*dxz/dxy;  */

/*   m[0]=yQP; m[1]=tQP; */
/*   m[2]=yQR; m[3]=tQR; dyt=det2(m); */
/*   m[0]=xQP; m[1]=tQP; */
/*   m[2]=xQR; m[3]=tQR; dxt=det2(m); */
/*   pT->t=pQ->t - (pT->x-pQ->x)*dyt/dxy + (pT->y-pQ->y)*dxt/dxy;  */

/*   m[0]=yQP; m[1]=lQP; */
/*   m[2]=yQR; m[3]=lQR; dyl=det2(m); */
/*   m[0]=xQP; m[1]=lQP; */
/*   m[2]=xQR; m[3]=lQR; dxl=det2(m); */
/*   pT->l=pQ->l - (pT->x-pQ->x)*dyl/dxy + (pT->y-pQ->y)*dxl/dxy;  */
/* } */


/* int istilted(point3D* pP, point3D* pQ, point3D* pR) */
/* { */
/*   float m[4]; */
/*   float xQP,yQP,xRP,yRP; */

/*   xQP=pQ->x-pP->x; yQP=pQ->y-pP->y; */
/*   xRP=pR->x-pP->x; yRP=pR->y-pP->y; */

/*   m[0]=xQP; m[1]=yQP; */
/*   m[2]=xRP; m[3]=yRP; */

/*   if (det2(m) !=0) */
/*     return YES; */
/*   else */
/*     return NO; */
/* } */

/* int intri(point3D* pM, point3D* pP, point3D* pQ, point3D* pR) */
/* { */
/*   int spmq,sqmr,srmp; */
/*   spmq=svp(pP,pM,pQ); sqmr=svp(pQ,pM,pR); srmp=svp(pR,pM,pP);  */

/*   if (spmq*sqmr==-1 || spmq*srmp==-1 || sqmr*srmp==-1) */
/*     return NO; */
/*   else */
/*     return YES; */
/* } */

/* void getdiag(point3D* pA, point3D* pB, point3D* pC, point3D* pD) */
/* { */
/*   int sacb,sadb; */
/*   point3D *pT; */

/*   sacb=svp(pA,pC,pB); sadb=svp(pA,pD,pB); */
/*   if (sacb==sadb) {  */
/*     pT=pB; pB=pC; pC=pT; */
/*     sacb=svp(pA,pC,pB); sadb=svp(pA,pD,pB); */
/*     if (sacb==sadb) { */
/*       pT=pB; pB=pD; pD=pT; */
/*     } */
/*   } */
/* } */

/* int svp(point3D* pP, point3D* pQ, point3D* pR) */
/* { */
/*   float m[4]; */
/*   float xQP,yQP,xQR,yQR; */
  
/*   xQP=pQ->x-pP->x; yQP=pQ->y-pP->y; */
/*   xQR=pQ->x-pR->x; yQR=pQ->y-pR->y; */

/*   m[0]=xQP; m[1]=yQP; */
/*   m[2]=xQR; m[3]=yQR; */

/*   if (det2(m)>=0) */
/*     return 1; */
/*   else */
/*     return -1; */
/* } */

float volum(point3D* pP, point3D* pQ, point3D* pR, point3D* pS)
{
  float m[9];
  float zRP,xRP,yRP, zQP,xQP,yQP, zSP,xSP,ySP;

  zRP=pR->z-pP->z; zQP=pQ->z-pP->z; zSP=pS->z-pP->z;
  xRP=pR->x-pP->x; xQP=pQ->x-pP->x; xSP=pS->x-pP->x;
  yRP=pR->y-pP->y; yQP=pQ->y-pP->y; ySP=pS->y-pP->y;

  m[0]=zRP; m[1]=zQP; m[2]=zSP;
  m[3]=xRP; m[4]=xQP; m[5]=xSP;
  m[6]=yRP; m[7]=yQP; m[8]=ySP;  

  return ABS(det3(m));
}

void get_times_slow(rays *pRR, triplets *ptrp,tetrahedra *ptet, 
		    cube *pTT, cube *pLL)
{
  int iR3,iT4;
  point3D *pA,*pB,*pC,*pD;
  float b2min,b2max,b3min,b3max,b1min,b1max;
  int n2min,n2max,n3min,n3max,n1min,n1max;
  int bi2,bi3,bi1;
  point3D M;
  int ai,aa,ia,bi,ba,ib,ci,ca,ic,di,da,id;
  
  for(iT4=0;iT4<ptet->nT4;iT4++) {
    for(iR3=0;iR3<ptrp->nR3;iR3++) {

      ai=ptet->T4[iT4][iR3].A_i; 
      aa=ptet->T4[iT4][iR3].A_a;
      ia=ptet->T4[iT4][iR3].ia;

      bi=ptet->T4[iT4][iR3].B_i;
      ba=ptet->T4[iT4][iR3].B_a;
      ib=ptet->T4[iT4][iR3].ib;

      ci=ptet->T4[iT4][iR3].C_i;
      ca=ptet->T4[iT4][iR3].C_a;
      ic=ptet->T4[iT4][iR3].ic;

      di=ptet->T4[iT4][iR3].D_i;
      da=ptet->T4[iT4][iR3].D_a;
      id=ptet->T4[iT4][iR3].id;

      pA=&pRR->c0[ai][aa][ia];
      pB=&pRR->c0[bi][ba][ib];
      pC=&pRR->c0[ci][ca][ic];
      pD=&pRR->c0[di][da][id];

      if (isT4(pA,pB,pC,pD)) {
	
	b2min=pA->x; b2max=pA->x;
	b2min=MIN(pB->x,b2min); b2min=MIN(pC->x,b2min); b2min=MIN(pD->x,b2min);
	b2max=MAX(pB->x,b2max); b2max=MAX(pC->x,b2max); b2max=MAX(pD->x,b2max);
	n2min=(int)(b2min-pTT->o2)/pTT->d2; n2max=(int)(b2max-pTT->o2)/pTT->d2+1;
	
	b3min=pA->y; b3max=pA->y; 
	b3min=MIN(pB->y,b3min); b3min=MIN(pC->y,b3min); b3min=MIN(pD->y,b3min);
	b3max=MAX(pB->y,b3max); b3max=MAX(pC->y,b3max); b3max=MAX(pD->y,b3max);
	n3min=(int)(b3min-pTT->o3)/pTT->d3; n3max=(int)(b3max-pTT->o3)/pTT->d3+1;
	
	b1min=pA->z; b1max=pA->z;
	b1min=MIN(pB->z,b1min); b1min=MIN(pC->z,b1min); b1min=MIN(pD->z,b1min);
	b1max=MAX(pB->z,b1max); b1max=MAX(pC->z,b1max); b1max=MAX(pD->z,b1max);
	n1min=(int)(b1min-pTT->o1)/pTT->d1; n1max=(int)(b1max-pTT->o1)/pTT->d1+1;
	
	n2min=(n2min<0)?0:n2min; n2max=(n2max<=pTT->n2)?n2max:pTT->n2;
	n3min=(n3min<0)?0:n3min; n3max=(n3max<=pTT->n3)?n3max:pTT->n3;
	n1min=(n1min<0)?0:n1min; n1max=(n1max<=pTT->n1)?n1max:pTT->n1;
	
	/* loop over the box around the tetrahedron */
	for(bi2=n2min;bi2<n2max;bi2++) {
	  for(bi3=n3min;bi3<n3max;bi3++) {
	    for(bi1=n1min;bi1<n1max;bi1++) {
	      M.x=pTT->o2+bi2*pTT->d2;
	      M.y=pTT->o3+bi3*pTT->d3;
	      M.z=pTT->o1+bi1*pTT->d1;

	      
	      if(isinT4(&M,pA,pB,pC,pD)==YES) {
		
		pA->t=pRR->ot+ia*pRR->st;
		pB->t=pRR->ot+ib*pRR->st;
		pC->t=pRR->ot+ic*pRR->st;
		pD->t=pRR->ot+id*pRR->st;
		
		getit(&M,pA,pB,pC,pD);
		if(pLL->c0[bi3][bi2][bi1]==0.0) {
		  pTT->c0[bi3][bi2][bi1]=M.t;
		  pLL->c0[bi3][bi2][bi1]=M.l;
		}
		else {
		  if (M.l<pLL->c0[bi3][bi2][bi1]) {
		    pTT->c0[bi3][bi2][bi1]=M.t;
		    pLL->c0[bi3][bi2][bi1]=M.l;
		  }
		}
	      }
	    }}}

      } /* end 'if isT4' */
    }}/* end loop iR3 and iT4 */
}
/* end 'if isinT4' */
/* end loop on the box around the T4 */

int isinT4(point3D* pM, point3D* pA, point3D* pB, point3D* pC, point3D* pD)
{
  /* check if M is inside (ABCD) */
  int sMA,sMB,sMC,sMD;

  sMA=sameside(pM,pA,   pB,pC,pD);
  sMB=sameside(pM,pB,pA,   pC,pD);
  sMC=sameside(pM,pC,pA,pB,   pD);
  sMD=sameside(pM,pD,pA,pB,pC   );
  
  if(sMA==YES && sMB==YES && sMC==YES && sMD==YES)
    return YES;
  else
    return NO;
}

int sameside(point3D* pM, point3D* pT, 
	     point3D* pX, point3D* pY, point3D* pZ)
{
  /* check if the apex and the interior point are on the same side of the base*/
  v3D XM,XY,XZ, XT;
  int sigM,sigT;

  XM.dx=pM->x-pX->x; XM.dy=pM->y-pX->y; XM.dz=pM->z-pX->z;
  XY.dx=pY->x-pX->x; XY.dy=pY->y-pX->y; XY.dz=pY->z-pX->z;
  XZ.dx=pZ->x-pX->x; XZ.dy=pZ->y-pX->y; XZ.dz=pZ->z-pX->z;
  XT.dx=pT->x-pX->x; XT.dy=pT->y-pX->y; XT.dz=pT->z-pX->z;

  sigM=sig(v_mp(&XM,&XY,&XZ));
  sigT=sig(v_mp(&XT,&XY,&XZ));

  if (sigM==sigT || sigM==0)  return YES;
  else                        return NO;
}

int isT4(point3D* pT, point3D* pX, point3D* pY, point3D* pZ)
{
  /* check if a T4 really is a T4 or is actually a planar object */
  v3D XY,XZ,XT;
  int sigT;
  XY.dx=pY->x-pX->x; XY.dy=pY->y-pX->y; XY.dz=pY->z-pX->z;
  XZ.dx=pZ->x-pX->x; XZ.dy=pZ->y-pX->y; XZ.dz=pZ->z-pX->z;
  XT.dx=pT->x-pX->x; XT.dy=pT->y-pX->y; XT.dz=pT->z-pX->z;

  sigT=sig(v_mp(&XT,&XY,&XZ));

  if (sigT!=0)  return YES;
  else          return NO;
}

void getit(point3D* pM, point3D* pA, point3D* pB, point3D* pC, point3D* pD)
{
  /* get the time for M which should be inside the T4 (ABCD) */
  float xMA,yMA,zMA, xBA,yBA,zBA,tBA,lBA, xCA,yCA,zCA,tCA,lCA, xDA,yDA,zDA,tDA,lDA;
  float m[9];
  float detxyz,detyzt,detxzt,detxyt, detyzl,detxzl,detxyl;

  xMA=pM->x-pA->x; yMA=pM->y-pA->y; zMA=pM->z-pA->z;
  xBA=pB->x-pA->x; yBA=pB->y-pA->y; zBA=pB->z-pA->z; tBA=pB->t-pA->t; lBA=pB->l-pA->l;
  xCA=pC->x-pA->x; yCA=pC->y-pA->y; zCA=pC->z-pA->z; tCA=pC->t-pA->t; lCA=pC->l-pA->l;
  xDA=pD->x-pA->x; yDA=pD->y-pA->y; zDA=pD->z-pA->z; tDA=pD->t-pA->t; lDA=pD->l-pA->l;

  m[0]=xBA; m[1]=yBA; m[2]=zBA;
  m[3]=xCA; m[4]=yCA; m[5]=zCA;
  m[6]=xDA; m[7]=yDA; m[8]=zDA;  detxyz=det3(m);

  m[0]=yBA; m[1]=zBA; m[2]=tBA;
  m[3]=yCA; m[4]=zCA; m[5]=tCA;
  m[6]=yDA; m[7]=zDA; m[8]=tDA;  detyzt=det3(m);
  m[0]=yBA; m[1]=zBA; m[2]=lBA;
  m[3]=yCA; m[4]=zCA; m[5]=lCA;
  m[6]=yDA; m[7]=zDA; m[8]=lDA;  detyzl=det3(m);
  
  m[0]=xBA; m[1]=zBA; m[2]=tBA;
  m[3]=xCA; m[4]=zCA; m[5]=tCA;
  m[6]=xDA; m[7]=zDA; m[8]=tDA;  detxzt=det3(m);
  m[0]=xBA; m[1]=zBA; m[2]=lBA;
  m[3]=xCA; m[4]=zCA; m[5]=lCA;
  m[6]=xDA; m[7]=zDA; m[8]=lDA;  detxzl=det3(m);

  m[0]=xBA; m[1]=yBA; m[2]=tBA;
  m[3]=xCA; m[4]=yCA; m[5]=tCA;
  m[6]=xDA; m[7]=yDA; m[8]=tDA;  detxyt=det3(m);
  m[0]=xBA; m[1]=yBA; m[2]=lBA;
  m[3]=xCA; m[4]=yCA; m[5]=lCA;
  m[6]=xDA; m[7]=yDA; m[8]=lDA;  detxyl=det3(m);

  pM->t=pA->t + (detyzt/detxyz)*xMA - (detxzt/detxyz)*yMA + (detxyt/detxyz)*zMA;
  pM->l=pA->l + (detyzl/detxyz)*xMA - (detxzl/detxyz)*yMA + (detxyl/detxyz)*zMA;
}


int integerup(float x, float ox, float dx)
{
  int n;
  if ( (int)((x-ox)/dx)==(x-ox)/dx )
    n=(x-ox)/dx;
  else
    n=(int)((x-ox)/dx)+1;
  return n;
}

int integerdn(float x, float ox, float dx)
{
  int n;
  n=(int)((x-ox)/dx);
  return n;
}

int intnn(float x, float ox, float dx)
{
  int n;
  n=(int)((x-ox)/dx+0.5);
  return n;
}

















