#include <math.h>
#include "struct.h"
#include "hwt_global.h"
#include "GL/glut.h"

GLfloat yangle = 0.0;
float x_length,y_length,z_length;
float x_view,y_view,z_view;
float view_radius;
float view_angle=0;
float zoom=2;
int raysplot=YES,tripletsplot=YES,triplet_number=0;
extern int sepxargc;
extern char **sepxargv;

rays *pRg;
cube *pTg;
triplets *ptrpg;
tetrahedra *ptetg;

void graph(rays*,triplets*,tetrahedra*,cube*);
void Keyboard(unsigned char key,int x,int y);
void Display();
void Reshape(int w,int h);
void InitGL(void);
void plot_vertex(rays*,cube*,point3D);
void plotxyz(float,float,float);
void plot_triplet(rays*,int);
void plot_tetraheda(rays*,triplets*,tetrahedra*,cube*,int);

void graph(rays*pRR, triplets *ptrp, tetrahedra *ptet, cube *pTT)
{
pTg=pTT;
pRg=pRR;
ptrpg=ptrp;
ptetg=ptet;

  InitGL();
  glutDisplayFunc(Display);
  glutReshapeFunc(Reshape);
  glutKeyboardFunc(Keyboard);
  glutMainLoop();
}

void InitGL(void)
{
  x_length=(pTg->n2+1)*pTg->d2; y_length=(pTg->n3+1)*pTg->d3;  z_length=(pTg->n1+1)*pTg->d1;
  view_radius=sqrt(pow(x_length,2)+pow(y_length,2)+pow(z_length,2));
  x_view=0.707*view_radius*cos(view_angle);
  y_view=      view_radius*sin(view_angle);
  z_view=0.707*view_radius*cos(view_angle);
  
  glutInit(&sepxargc, sepxargv);
  glutInitDisplayMode (GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
  glutInitWindowSize(500,500);
  glutInitWindowPosition(0.0,0.0);
  glutCreateWindow("Tessalation");
  
  glClearColor (0.0, 0.0, 0.0, 0.0);
  glShadeModel (GL_FLAT);
}

void Reshape (int w, int h)
{
  glViewport (0, 0, (GLsizei) w, (GLsizei) h);
  glMatrixMode (GL_PROJECTION);
  glLoadIdentity();
  gluPerspective (30.0,1.0,4*z_view,0.0);
  glMatrixMode (GL_MODELVIEW);
}

void Display(void)
{
  int ii,it,ia;

  glClear (GL_COLOR_BUFFER_BIT);
  glPushMatrix ();
  
  gluLookAt (zoom*x_view,zoom*y_view,zoom*z_view,
	     0.0, 0.0, 0.0,
	     0.0, 1.0, 0.0);
  
  /* ROTATE FROM HERE DOWN */
  glRotatef (yangle,0.0,1.0,0.0);  
  /*mobile axes*/
  glBegin (GL_LINES);
  glColor3f  (0.0, 1.0, 1.0);
  
  plotxyz(x_length,0       ,0       ); plotxyz(x_length,z_length,0       );  
  plotxyz(0       ,0       ,y_length); plotxyz(0       ,z_length,y_length);
  plotxyz(x_length,0       ,y_length); plotxyz(x_length,z_length,y_length);
  
  plotxyz(x_length,0       ,0       ); plotxyz(x_length,       0,y_length);
  plotxyz(x_length,z_length,0       ); plotxyz(x_length,z_length,y_length);
  plotxyz(0       ,z_length,0       ); plotxyz(0,       z_length,y_length);
  
  plotxyz(0       ,z_length,y_length); plotxyz(x_length,z_length,y_length);
  plotxyz(0       ,       0,y_length); plotxyz(x_length,0       ,y_length);
  plotxyz(0       ,z_length,0       ); plotxyz(x_length,z_length,0       );
  
  /* Oz */  glColor3f (1.0, 0.0, 0.0);
  plotxyz(0,0,0); plotxyz(0,z_length,0);
  
  /* Oy */  glColor3f (0.0, 1.0, 0.0);
  plotxyz(0,0,0); plotxyz(0,0,y_length);
  
  /* Ox */  glColor3f (0.0, 0.0, 1.0);
  plotxyz(0,0,0); plotxyz(x_length,0,0); 

  glEnd();

  if(raysplot) {
      glBegin (GL_LINES);
      glColor3f(1.0,0.0,0.0);
      for(ii=0;ii<pRg->ni;ii++) {
	for(ia=0;ia<pRg->na;ia++) {
	  for(it=0;it<pRg->nt-1;it++) {
	    plot_vertex(pRg,pTg,pRg->c0[ii][ia][it]);
	    plot_vertex(pRg,pTg,pRg->c0[ii][ia][it+1]);

	  }}}
      glEnd();
  }/* end rays */
  
  if(tripletsplot) {
    plot_tetraheda(pRg,ptrpg,ptetg,pTg,triplet_number);
    plot_triplet(pRg,triplet_number);
  }/* end triplets */

  glPopMatrix ();
  glutSwapBuffers ();
}

void Keyboard( unsigned char key, int x, int y )
{
  switch (key) {
  case 27:  /* ESCAPE */
    /*    cleanup ();*/
    exit(0);
    break;
  case 'y':
    yangle += 5.0;
    if (yangle > 360.) yangle = yangle - 360.;
    glutPostRedisplay (); 
    break;
  case 'Y':
    yangle -= 5.0;
    if (yangle < 0.) yangle = yangle + 360.;
    glutPostRedisplay (); 
    break;
  case 'u':
    view_angle+= PI/24;
    if (view_angle > 11*PI/24) view_angle=11*PI/24;
    x_view=0.707*view_radius*cos(view_angle);
    y_view=      view_radius*sin(view_angle);
    z_view=0.707*view_radius*cos(view_angle);
    glutPostRedisplay ();
    break;
  case 'd':
    view_angle-= PI/24;
    if (view_angle < -11*PI/24) view_angle=-11*PI/24;
    x_view=0.707*view_radius*cos(view_angle);
    y_view=      view_radius*sin(view_angle);
    z_view=0.707*view_radius*cos(view_angle);
    glutPostRedisplay ();
    break;
  case 'i':
    zoom -= 0.1;
    if (zoom < 0.1) zoom = 0.1;
    glutPostRedisplay (); 
    break;
  case 'o':
    zoom += 0.1;
    if (zoom > 10.0) zoom = 10.0;
    glutPostRedisplay (); 
    break;
  case 'r':
    if(raysplot==YES) raysplot=NO;
    else raysplot=YES;
    glutPostRedisplay (); 
    break;
  case 't':
    if(tripletsplot==YES) tripletsplot=NO;
    else tripletsplot=YES;
    glutPostRedisplay (); 
    break;
  case ']':
    triplet_number+=1;
    if (triplet_number>ptrpg->nR3-1) triplet_number=ptrpg->nR3-1;
    glutPostRedisplay ();
    break;
  case '[':
    triplet_number-=1;
    if (triplet_number<0) triplet_number=0;
    glutPostRedisplay ();
    break;
  default:
    break;
  }
}

void plot_vertex(rays *pRR, cube *pTT,point3D P)
{
  float min2;
  float min3;
  float min1;
 
  min2=pTT->o2;
  min3=pTT->o3;
  min1=pTT->o1;
  plotxyz(P.x-min2,P.z-min1,P.y-min3);
}

void plotxyz(float x, float z, float y)
{
  float xtr=x_length/2,ytr=y_length/2,ztr=z_length/2;
  glVertex3f (x-xtr,-(z-ztr),-(y-ytr));
}

void plot_triplet(rays* pRR, int triplet_number)
{
  int ii,it,ia;

  glBegin (GL_LINES);
  glColor3f(0.0,1.0,0.0);

  get_R3(pRg,ptrpg,triplet_number);

  for(it=0;it<pRg->nt-1;it++) {
    /* ray A */
    plot_vertex(pRR,pTg,ptrpg->rayA.point[it]);
    plot_vertex(pRR,pTg,ptrpg->rayA.point[it+1]);
    
    /* ray B */
    plot_vertex(pRR,pTg,ptrpg->rayB.point[it]);
    plot_vertex(pRR,pTg,ptrpg->rayB.point[it+1]);
    
    /* ray C */
    plot_vertex(pRR,pTg,ptrpg->rayC.point[it]);
    plot_vertex(pRR,pTg,ptrpg->rayC.point[it+1]);
  }
  glEnd();
}

void plot_tetraheda(rays *pRR,triplets *ptrp, tetrahedra *ptet, cube *pTT, int iR3)
{
  int iT4;
  point3D A,B,C,D;
  int ai,aa,ia,bi,ba,ib,ci,ca,ic,di,da,id;
  
  glBegin (GL_LINES);
  glColor3f(1.0,1.0,1.0);
  for(iT4=0;iT4<ptetg->nT4;iT4++) {

    ai=ptet->T4[iT4][iR3].A_i;
    aa=ptet->T4[iT4][iR3].A_a;
    ia=ptet->T4[iT4][iR3].ia;

    bi=ptet->T4[iT4][iR3].B_i;
    ba=ptet->T4[iT4][iR3].B_a;
    ib=ptet->T4[iT4][iR3].ib;
    
    ci=ptet->T4[iT4][iR3].C_i;
    ca=ptet->T4[iT4][iR3].C_a;
    ic=ptet->T4[iT4][iR3].ic;

    di=ptet->T4[iT4][iR3].D_i;
    da=ptet->T4[iT4][iR3].D_a;
    id=ptet->T4[iT4][iR3].id;

    A=pRR->c0[ai][aa][ia];
    B=pRR->c0[bi][ba][ib];
    C=pRR->c0[ci][ca][ic];
    D=pRR->c0[di][da][id];

    plot_vertex(pRR,pTT,A); plot_vertex(pRR,pTT,B);
    plot_vertex(pRR,pTT,A); plot_vertex(pRR,pTT,C);
    plot_vertex(pRR,pTT,A); plot_vertex(pRR,pTT,D);
    plot_vertex(pRR,pTT,B); plot_vertex(pRR,pTT,C);
    plot_vertex(pRR,pTT,B); plot_vertex(pRR,pTT,D);
    plot_vertex(pRR,pTT,C); plot_vertex(pRR,pTT,D);
  }
  glEnd();
}



