program Genvelmodel
! Program to create the velocity model at page 385 in: Hatchell, P. J., 
! "Fault Whispers, Transmission distortions on prestack seismic reflection 
! data,"Geophysics, Vol 65, No. 2, pp. 377 - 389, March - April 2000.

! Input: v1, v2 and EITHER nx, nz, x1, x2, z1, z2, z3 (setsampling=n) 
!                   OR     dx, dz,nx1,nx2,nz1,nz2,nz3 (setsampling=y)
! Output: (nz,nx) sized velocity model.
! Parameters and vel model: described in the drawing below
!
! <--------x1--------><--------x2-------->
! ^                   
! |
! |
! |
! z1        v1                 v1
! |
! |
! |
! v
! ____________________
! ^                  |
! z2        v2       |         v1
! v                  |
! --------------------
! ^                   
! |
! |
! |
! z3        v1                 v1
! |
! |
! |
! v
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  use sep
  implicit none

  logical :: setsampling
  integer :: nx, nz, nx1, nx2, nz1, nz2, nz3, ix, iz, esize
  real    :: v1, v2, x1, x2, z1, z2, z3, dx, dz

  call sep_init ()
  call from_param ("v1", v1, 1830. )
  call from_param ("v2", v2, 1647. )

  call from_param ("setsampling", setsampling, .true.)
  if(setsampling) then
     call from_param ("dx" ,  dx,  12.)
     call from_param ("dz" ,  dz,  12.)
     call from_param ("nx1", nx1, 256 )
     call from_param ("nx2", nx2, 256 )
     call from_param ("nz1", nz1, 256 )
     call from_param ("nz2", nz2,  13 )
     call from_param ("nz3", nz3, 243 )
     nx = nx1 + nx2
     nz = nz1 + nz2 + nz3
  else
     call from_param ("nx", nx,  512  )
     call from_param ("nz", nz,  512  )
     call from_param ("x1", x1, 3000. )
     call from_param ("x2", x2, 3000. )
     call from_param ("z1", z1, 3050. )
     call from_param ("z2", z2,  152.5)
     call from_param ("z3", z3, 2897.5)
     dx = (x1 + x2    )/(nx-1)
     dz = (z1 + z2 +z3)/(nz-1)
     nx1 = nint(x1/dx)
     nz1 = nint(z1/dz)
     nz2 = nint(z2/dz)
  end if

  call to_history ("n1",  nz)
  call to_history ("n2",  nx)
  call to_history ("o1",  0.)
  call to_history ("o2",  0.)
  call to_history ("d1",  dz)
  call to_history ("d2",  dx)

  esize=4
  call to_history ("esize",esize)
  call sep_close ()

  xloop:do ix=1,nx
     zloop:do iz=1,nz
        if(iz>nz1.and.iz<nz1+nz2.and.ix<nx1) then
           call srite('out', v2, esize)
        else
           call srite('out', v1, esize)
        end if
     end do zloop
  end do xloop

end program Genvelmodel
