program Linv2

  !-improved version of Zhang's correction----.
  !pd(x,z=0;w) = -0.5/lambda*delta(x-xs)                           !

  use sep, only: sep_init, from_param, sep_close, sep_read, sep_write
  use ooio, only: axis, readaxis
  use fft_data_mod, only: fft_type, init_fft_1d, init_rfft_1d, fft_1d_data, rfft_1d_data, clean_fft

  implicit none
  real   , allocatable, dimension (:,:) :: U ! input wavefield, 2-D, offset=0
  complex, allocatable, dimension (:,:) :: C ! wavefield in Fourier domain
  type(axis)                            ::  x, t, w, k
  integer                               :: iw,ik, ix
  type(fft_type)                        :: fft_1d_id, rft_1d_id
  real                                  :: v0, sqa !sqrt(kz)
  complex, dimension(:), allocatable    :: all_freq
  real,    dimension(:), allocatable    :: G
  complex, parameter                    :: i = cmplx(0.,1.)
  real,    parameter                    :: pi=3.1415926535898

  call sep_init()

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  ! Read/write axes
  t = readaxis(1, chk_even=.true.)
  if(t%o/=0) call seperr('For migration t must start at zero (o1=0 in input)')

  x = readaxis(2, chk_even=.true.)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  ! Velocity stuff
  call sep_close()
  call from_param('v0', v0, 1000.)
  if(v0<=0.) call seperr('zero or neg vel!')
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  ! Temporal frequency axis
  w%n = t%n/2
  w%d = 1./(t%n*t%d)
  w%o = w%d ! no zero frequency component
  allocate(w%all(w%n)); w%all=0.
  do iw = 1, w%n
     w%all(iw) = w%o  + (iw - 1) * w%d
  end do
  w%all = 2*pi * w%all
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  ! Fourier transform stuff
  rft_1d_id = init_rfft_1d(t%n)
  fft_1d_id = init_fft_1d(x%n)
!!!!!!!!!!!!!!!!!!!!!!!!!!!
  ! Spatial frequency axis
  k%n = x%n
  allocate(k%all(k%n)); k%all=0.
  k%d = 1./(k%n*x%d)
  k%o = 1./(2*x%d)-k%d !k does not increase uniformly!
  do ik=1, k%n
     k%all(ik) = mod(k%o+(ik-1)*k%d,1./x%d)-k%o
  end do

  k%all = 2*pi * k%all

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  ! Input data
  allocate (U(t%n,x%n))
  call sep_read(U)
  allocate (C(w%n,x%n)); C=0! nt/2 frequencies, no zero freq

  allocate(all_freq(w%n+1))
  do ix=1, x%n
     call rfft_1d_data(rft_1d_id, -1, t%n, U(:,ix), all_freq)
     C(:,ix) = all_freq(2:)
  end do

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  do iw = 1, w%n

     call fft_1d_data(fft_1d_id, -1, k%n, C(iw,:))

     do ik = 1, k%n
        sqa = w%all(iw)**2 / v0**2 - k%all(ik)**2
        avoid_evan:if ( sqa > 0 ) then
           ! 1/(2 Lambda)
           C(iw,ik) = C(iw,ik) / (2.*i*sqrt(sqa))
        else
           C(iw,ik) = 0.
        end if avoid_evan
     end do

     call fft_1d_data(fft_1d_id, 1, k%n, C(iw,:))

  end do

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  ! Bye.
  call clean_fft(fft_1d_id)
  do ix = 1, x%n
     all_freq(2:) = C(:,ix)
     call rfft_1d_data(rft_1d_id, 1, t%n, U(:,ix), all_freq)
  end do
  call clean_fft(rft_1d_id)
  deallocate(C, all_freq)
  call sep_write(U)
  deallocate(U)
end program Linv2
