program Wacon
  ! WAvefield CONtinuation by finite-difference or split-step
  ! 1. migration of receiver data (continuation of upward going wavefield): dowcon_as_shot=n
  ! 2. dowcon a shot, look at last wavefield (NOT migrated image): dowcon_as_shot=y
  ! For any of these options, can specify: zhang=y/n splitstep=y/n, want_freq_cuben, or want_freq_cube=y freq_cube=<filename.H>

  use sep, only: sep_init, from_param, sep_close, sep_read, sep_write, to_history
  use ooio, only: axis, readcharpar, readaxis, writeaxis, chk_dim_corresp
  use wacon_mod, only: wacon_init, wacon_mixdom, wacon_findif, clean_wacon, rft_fwd_axis1, rft_bck_axis1

  implicit none
  real   , allocatable, dimension (:,:) :: U ! input wavefield, 2-D, offset=0
  real   , allocatable, dimension (:,:) :: v ! interval velocity field
  complex, allocatable, dimension (:,:) :: C ! wavefield in Fourier domain
  real   , allocatable, dimension (:)   :: vz ! vertical velocity gradient
  integer                               :: ix,iz
  type(axis)                            ::  x, z, t, w
  logical                               :: dowcon_as_shot
  logical                               :: splitstep, zhang, want_freq_cube
  character(len=6)                      :: boundcond ! for finite difference
  ! For imaging condition by wavefield division:
!  real, parameter                       :: epsilon = 0.01
 ! real                                  :: erms

  call sep_init()
  call from_param('want_freq_cube', want_freq_cube, .false.)
  call from_param('dowcon_as_shot', dowcon_as_shot, .false.)
  call from_param('splitstep'     , splitstep     , .true. )
  call from_param('zhang'         , zhang         , .false.)

  if(.not.splitstep) boundcond = readcharpar('boundcond',(/'reflec','zslope'/), 2)

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  ! Read/write axes
  t = readaxis(1, chk_even=.true.)
  if(t%o/=0) call seperr('For migration t must start at zero (o1=0 in input)')

  z = readaxis(1, from='vel')
  x = readaxis(2, chk_even=.true.)

  ! if(dowcon_as_shot)  out has same dim as in
  if(.not.dowcon_as_shot.and..not.want_freq_cube) then
     call writeaxis(x, 1, axisname='x')
     call writeaxis(z, 2, axisname='z')
  end if

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  w = wacon_init(dowcon_as_shot, splitstep, zhang, x, z, t)

  if(want_freq_cube) then
     call writeaxis(w, 1, axisname='w', filetag='freq_cube')
     call writeaxis(x, 2, axisname='x', filetag='freq_cube')
     call writeaxis(z, 3, axisname='z', filetag='freq_cube')
     call to_history('esize', 8, 'freq_cube')
  end if

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  ! Velocity stuff
  call chk_dim_corresp('vel' ,(/z, x/))
  call sep_close()
  allocate( v(z%n,x%n))
  call sep_read(v,'vel')
  if(any(v<=0.)) call seperr('zero or neg value in the velocity model!')
  if(.not.dowcon_as_shot.and..not.want_freq_cube) v = v/2.

  if(zhang) then
     allocate(vz(x%n)); vz=0 !vertical velocity gradient
  end if
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  ! Input data
  allocate (U(t%n,x%n))
  call sep_read(U)
  allocate (C(w%n,x%n)); C=0! nt/2 frequencies, no zero freq
  C = rft_fwd_axis1(U)
  deallocate(U)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  ! Wavefield continuation:
  depth:do iz = 1, z%n
     if(zhang) then
        grad:if(iz==1) then
           vz = ( v(2,:) - v(1,:) ) / z%d
        else if(iz==z%n) then
           vz = ( v(z%n,:) - v(z%n-1,:) ) / z%d
        else
           vz = ( v(iz+1,:) - v(iz-1,:) ) / (2. * z%d)
        end if grad
        !vz = -vz
        if(     splitstep) call wacon_mixdom(C, v(iz,:),            vz)
        if(.not.splitstep) call wacon_findif(C, v(iz,:), boundcond, vz)
     else
        if(     splitstep) call wacon_mixdom(C, v(iz,:))
        if(.not.splitstep) call wacon_findif(C, v(iz,:), boundcond)
     end if

     if(.not.dowcon_as_shot.and..not.want_freq_cube) call sep_write(sum(real(C),1)) ! imag cond sum w
     if(want_freq_cube) call sep_write(C,'freq_cube')

     write(0,*) iz*100./z%n, '% downward continuation'
  end do depth

  deallocate(v)
  if(zhang) deallocate(vz)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  ! Bye.
  if(dowcon_as_shot) then
     allocate (U(t%n,x%n))
     U = rft_bck_axis1(C)
     call sep_write(U)
     deallocate(U)
  end if

  deallocate(C)
  call clean_wacon(splitstep)
end program Wacon
