program Zodocon
  ! Programmed after Paul's tutorial in SEP105, eq. 12. ATTENTION AT 2*PI!!
  use sep
  use ooio
  use myfft_mod
  use fft_data_mod
  use bodocon_mod
  implicit none

  real   , dimension(:,:), allocatable :: U, s
  complex, dimension(:,:), allocatable :: C, Czero, Ctmp
  real                                 :: s0, eta
  type(axis)                           :: t, x, z, w, k
  type(fft_id)                         :: id
  type(fft_type)                       :: fft_1d_id
  integer                              :: iz
  logical                              :: background_only, nonlinear

  call sep_init()
  call from_param('background_only',background_only,.false.)
  call from_param('nonlinear'      ,nonlinear      ,.false.)
  call from_param('eta'            ,eta            ,0.001  )

  t = readaxis(1, chk_even=.true.)
  x = readaxis(2, chk_even=.true.)
  if(background_only) then
     z%n = 500; z%d = 12.
     call from_param('v0', s0, 1830.)
     if(s0==0.) call seperr('average velocity is zero!!'); s0=1/s0
  else
     z = readaxis(1, from='v')
     call chk_dim_corresp('v',(/z, x/))
     allocate(s(z%n,x%n))
     call sep_read(s,'v')
     call from_param('v0', s0, sum(s)/(z%n*x%n))
     if(s0==0.) call seperr('average velocity is zero!!'); s0=1/s0
     if(any(s==0.)) call seperr('zero value in velocity function!')
     s=1./s-s0; !slownesses, and s will be delta s 
  end if
  call sep_close()
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  ! Init stuff for Fourier transforms
  id = init_my_2d_fft(t%n, x%n)
  fft_1d_id = init_fft_1d(x%n)
  ! Time frequency, done with RFFTW
  w%n = t%n/2+1
  w%d = 1./(t%n*t%d)
  w%o = 0.
  ! Spatial frequency, done with FFTW
  k%n = x%n
  k%d = 1./(x%n*x%d)
  k%o = 0.5/x%d-k%d !k%v will be mod(k%o+(ik-1)*k%d,1./x%d)-k%o
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  ! Read 2D, zero-offset seismogram
write(0,*) 'before allocate'
  allocate(U(t%n,x%n)); U=0.
  call sep_read(U)
  allocate(C(w%n,k%n)); C=0.
write(0,*) 'after allocate'

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
write(0,*) 'before forward FFT'
  C = fft_2d_fwd(U, id)
write(0,*) 'after forward FFT'
  bg_only:if(background_only) then
write(0,*) 'bg_only, before t_operator'
     call t_operator(C,w,k,z%d,x%d,s0,eta,z%n)
write(0,*) 'bg_only, after t_operator'
  else
write(0,*) 'BEGIN NOT bg_only. Before first t_operator'
     call t_operator(C,w,k,z%d,x%d,s0,eta,255)
write(0,*) 'After first t_operator'
!!!!!!!!!!!True meat, general case, begins here
write(0,*) 'Before another allocation'
     allocate(Czero(w%n,k%n))
write(0,*) 'After another allocation'
     nonlin:if(nonlinear) then
        Czero=0. ! Will keep current background wavefield in it

        slab:do iz=256,268
write(0,*) 'Nonlinear case, inside slab, step ', iz-255
write(0,*) '               Before t_op'
           call t_operator(C,w,k,z%d,x%d,s0,eta,1)
write(0,*) '               After  t_op'
           Czero = C
write(0,*) '               Before delta_s_op'
           call delta_s_operator(C, (s(iz,:)+s(iz+1,:))/2., fft_1d_id)
write(0,*) '               After  delta_s_op, before s_op'
           call s_operator(C,w,k,z%d,x%d,s0,eta)
           C = C + Czero
        end do slab

     else
        Czero = C !Background wavefield. For iz=1, should be fft(U)
        C     = 0.! C will be total scattered wavefield
write(0,*) 'Linear case, before another allocation'
        allocate(Ctmp(w%n,k%n)); Ctmp  = 0. !current scat wavefield
write(0,*) 'Linear case, after another allocation'
        in_slab:do iz=256,268
write(0,*) 'Linear case, inside slab, step ', iz-255
           call t_operator(Czero,w,k,z%d,x%d,s0,eta,1)
           call t_operator(C    ,w,k,z%d,x%d,s0,eta,1)
           Ctmp = Czero
           call delta_s_operator(Ctmp, (s(iz,:)+s(iz+1,:))/2., fft_1d_id)
           call s_operator(Ctmp,w,k,z%d,x%d,s0,eta)
           C = C + Ctmp
        end do in_slab
        deallocate(Ctmp)
     end if nonlin
     deallocate(Czero, s)
write(0,*) 'Not bg_only, end of main work'
!!!!!!!!!!!True meat, general case, ends here
     call t_operator(C,w,k,z%d,x%d,s0,eta,z%n-268)

  end if bg_only
write(0,*) 'Before reverse FFT'
  U = fft_2d_inv(C, id)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  deallocate(C)
  call sep_write(U)
  deallocate(U)
  call clean_fft(fft_1d_id)
  call clean_my_2d_fft(id)
end program Zodocon
