module bodocon_mod
  use ooio
  use fft_data_mod
  implicit none
  public :: t_operator, s_operator
  real, parameter:: pi2=6.283185307
contains

  subroutine t_operator(C,w,k,dz,dx,s0,eta,nsteps)
    complex, dimension(:,:), intent(inout) :: C
    type(axis),              intent(inout) :: w, k
    real,                    intent(in   ) :: dz, dx, eta, s0
    integer,                 intent(in   ) :: nsteps
    integer                                :: iw, ik

    do iw=1, w%n
       w%v = pi2 * (w%o + (iw-1)*w%d)
       do ik=1, k%n
          k%v = pi2 * (mod(k%o+(ik-1)*k%d,1./dx)-k%o)
          C(iw,ik) = C(iw,ik) * exp(dz*cmplx(0.,1.)*nsteps*sqrt((w%v*s0)**2-(k%v*cmplx(1.,-eta))**2))
       end do
    end do
  end subroutine t_operator
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  subroutine s_operator(C,w,k,dz,dx,s0,eta)
    complex, dimension(:,:), intent(inout) :: C
    type(axis),              intent(inout) :: w, k
    real,                    intent(in   ) :: dz, dx, eta, s0
    integer                                :: iw, ik

    do iw=1, w%n
       w%v = pi2 * (w%o + (iw-1)*w%d)
       do ik=1, k%n
          k%v = pi2 * (mod(k%o+(ik-1)*k%d,1./dx)-k%o)
          if(iw/=1.and.ik/=1) C(iw,ik) = C(iw,ik) * w%v*w%v/sqrt((w%v*s0)**2-(k%v*cmplx(1.,-eta))**2)
       end do
    end do
    C = C * cmplx(0.,dz*s0)
  end subroutine s_operator
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  subroutine delta_s_operator(C, sline, fft_1d_id)
    complex, dimension(:,:), intent(inout) :: C
    real,    dimension(:  ), intent(in   ) :: sline
    type(fft_type)         , intent(in   ) :: fft_1d_id
    integer                                :: nw, nk, iw

    nw = size(C,1)
    nk = size(C,2)

    do iw = 1, nw
       call fft_1d_data(fft_1d_id,FFTW_BACKWARD, nk, C(iw,:))
       C(iw,:) = C(iw,:) * sline
       call fft_1d_data(fft_1d_id,FFTW_FORWARD , nk, C(iw,:))
    end do
  end subroutine delta_s_operator
  end module bodocon_mod
