module myfft_mod

  use fft_data_mod

  implicit none

  type fft_id
     type(fft_type) :: fft_1d, rfft_1d
  end type fft_id

  private
  public :: fft_id, init_my_2d_fft, clean_my_2d_fft, fft_2d_fwd, fft_2d_inv, rft_1d_fwd, rft_1d_inv

contains

  function fft_2d_fwd(slice,id)
    real, dimension(:,:), intent(in)                    :: slice
    type(fft_id)        , intent(in)                    :: id
    complex, dimension(size(slice,1)/2+1,size(slice,2)) :: fft_2d_fwd

    integer :: ik, iw, nk, nw , nt

    nt = size(slice,1)
    nw = nt/2+1
    nk = size(slice,2)

    do ik = 1, nk
       call rfft_1d_data(id%rfft_1d,FFTW_REAL_TO_COMPLEX,nt,slice(:,ik),fft_2d_fwd(:,ik))
    end do

    do iw = 1, nw
       call fft_1d_data(id%fft_1d,FFTW_FORWARD, nk, fft_2d_fwd(iw,:))
    end do
  end function fft_2d_fwd
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  function rft_1d_fwd(slice,rfft_id)
    real, dimension(:,:), intent(in)                    :: slice
    type(fft_type)      , intent(in)                    :: rfft_id
    complex, dimension(size(slice,1)/2+1,size(slice,2)) :: rft_1d_fwd

    integer :: ik, nk, nt

    nt = size(slice,1)
    nk = size(slice,2)

    do ik = 1, nk
       call rfft_1d_data(rfft_id,FFTW_REAL_TO_COMPLEX,nt,slice(:,ik),rft_1d_fwd(:,ik))
    end do

  end function rft_1d_fwd
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  function rft_1d_inv(slice,rfft_id)
    complex, dimension(:,:), intent(in)                :: slice
    type(fft_type)         , intent(in)                :: rfft_id
    real, dimension(2*(size(slice,1)-1),size(slice,2)) :: rft_1d_inv

    integer :: ik, nk, nt

    nt = 2*(size(slice,1)-1)
    nk = size(slice,2)

    do ik = 1, nk
       call rfft_1d_data(rfft_id,FFTW_COMPLEX_TO_REAL,nt,rft_1d_inv(:,ik),slice(:,ik))
    end do

  end function rft_1d_inv
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  function fft_2d_inv(slice,id)
    complex, dimension(:,:), intent(in)                :: slice
    type(fft_id)           , intent(in)                :: id
    real, dimension(2*(size(slice,1)-1),size(slice,2)) :: fft_2d_inv

    integer :: ik, iw, nk, nw , nt

    nw = size(slice,1)
    nt = 2*(nw-1)
    nk = size(slice,2)

    do iw = 1, nw
       call fft_1d_data(id%fft_1d,FFTW_BACKWARD, nk, slice(iw,:))
    end do

    do ik = 1, nk
       call rfft_1d_data(id%rfft_1d,FFTW_COMPLEX_TO_REAL,nt,fft_2d_inv(:,ik),slice(:,ik))
    end do

  end function fft_2d_inv
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  function init_my_2d_fft(n_rfft,n_fft)
    integer, intent(in) :: n_rfft,n_fft
    type(fft_id)        :: init_my_2d_fft

    init_my_2d_fft%rfft_1d = init_rfft_1d(n_rfft)
    init_my_2d_fft%fft_1d  = init_fft_1d(n_fft)
  end function init_my_2d_fft
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  subroutine clean_my_2d_fft(id)
    type(fft_id)           , intent(inout) :: id

    call clean_rfft(id%rfft_1d)
    call clean_fft(id%fft_1d)

  end subroutine clean_my_2d_fft
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
end module myfft_mod
