module ooio
  ! object-oriented input-output
  ! This version eliminated writeaxis bug, ooi2 still has it
  use sep

  type axis
     integer :: n
     real    :: o, d, v !v=for the value of the variable (typical:v=(i-1)*d+o)
     real, dimension(:), allocatable:: all
  end type axis

  private
  public :: axis, readaxis, chk_dim_corresp, writeaxis, readcharpar

contains
!---------------------------------------------------------------
  subroutine chk_dim_corresp(filetag,dimarray)
    !call chk_dim_corresp('vel',(/z, x/)) in same order as dims of velfile
    character(len=*)        , intent(in):: filetag
    type(axis), dimension(:), intent(in):: dimarray

    integer                     :: n, i
    real                        :: r
    character(len=9):: stringbank 
    stringbank = '123456789'

    if(.not.exist_file(filetag)) call seperr('aux file does not exist')

     do i=1, size(dimarray)
        write(0,*) 'Checking dimension ',i,' correspondence between ', &
        filetag,' file and input seismic file'
        call from_aux(filetag,'n'//stringbank(i:i),n)
        if(n/=dimarray(i)%n) call seperr("velfile needs have same n's as inp")
        call from_aux(filetag,'o'//stringbank(i:i),r)
        if(r/=dimarray(i)%o) call seperr("velfile needs have same o's as inp")
        call from_aux(filetag,'d'//stringbank(i:i),r)
        if(r/=dimarray(i)%d) call seperr("velfile needs have same d's as inp")
        write(0,*) 'OK'
     end do

  end subroutine chk_dim_corresp
!----------------------------------------------------------
  function readaxis(axisnr, axisname, from, ndefault, odefault, ddefault, chk_even)
    integer         , intent(in), optional :: axisnr, ndefault
    character(len=*), intent(in), optional :: axisname
    character(len=*), intent(in), optional :: from
    real            , intent(in), optional :: odefault, ddefault
    logical         , intent(in), optional :: chk_even

    type(axis)                             :: readaxis

    integer                                :: n
    real                                   :: o, d
    character(len=9)                       :: stringbank 
    character(len=max(1,len(axisname)))    :: axhandle 
    
    if(present(axisnr)) then
       if(present(axisname)) call seperr('readaxis: not both axis name and #')
       if(axisnr<1.or.axisnr>9) call seperr('readaxis: axis# must be from 1-9')
       stringbank = '123456789'
       axhandle   = stringbank(axisnr:axisnr)
    else if(present(axisname)) then
       axhandle   = axisname
    else
       call seperr('readaxis: specify an axis name or number')
    end if

    if_from:if(present(from)) then
       if_par:if(from=='par') then
          if(present(ndefault)) then
             call from_param('n'//axhandle, readaxis%n, ndefault)
          else
             call from_param('n'//axhandle, readaxis%n)
          end if
          if(present(odefault)) then
             call from_param('o'//axhandle, readaxis%o, odefault)
          else
             call from_param('o'//axhandle, readaxis%o)
          end if
          if(present(ddefault)) then
             call from_param('d'//axhandle, readaxis%d, ddefault)
          else
             call from_param('d'//axhandle, readaxis%d)
          end if
       else if(from=='in') then
          if(present(ndefault)) then
             call from_history('n'//axhandle, readaxis%n, ndefault)
          else
             call from_history('n'//axhandle, readaxis%n)
          end if
          if(present(odefault)) then
             call from_history('o'//axhandle, readaxis%o, odefault)
          else
             call from_history('o'//axhandle, readaxis%o)
          end if
          if(present(ddefault)) then
             call from_history('d'//axhandle, readaxis%d, ddefault)
          else
             call from_history('d'//axhandle, readaxis%d)
          end if
       else
          if(.not.exist_file(from)) call seperr('Aux file does not exist')
          if(present(ndefault)) then
             call from_aux(from,'n'//axhandle, readaxis%n, ndefault)
          else
             call from_aux(from,'n'//axhandle, readaxis%n)
          end if
          if(present(odefault)) then
             call from_aux(from,'o'//axhandle, readaxis%o, odefault)
          else
             call from_aux(from,'o'//axhandle, readaxis%o)
          end if
          if(present(ddefault)) then
             call from_aux(from,'d'//axhandle, readaxis%d, ddefault)
          else
             call from_aux(from,'d'//axhandle, readaxis%d)
          end if
       end if if_par
    else
       if(present(ndefault)) then
          call from_history('n'//axhandle, readaxis%n, ndefault)
       else
          call from_history('n'//axhandle, readaxis%n)
       end if
       if(present(odefault)) then
          call from_history('o'//axhandle, readaxis%o, odefault)
       else
          call from_history('o'//axhandle, readaxis%o)
       end if
       if(present(ddefault)) then
          call from_history('d'//axhandle, readaxis%d, ddefault)
       else
          call from_history('d'//axhandle, readaxis%d)
       end if
    end if if_from

    if(present(chk_even)) then
       if(mod(readaxis%n,2)/=0) call seperr('Axis length must be even')
    end if
  end function readaxis
!------------------------------------------------------------------------
  subroutine writeaxis(axisvar, axisnr, filetag, axisname)
    !Filetag is not the file name! Example: <in.H Prog.x out2=abc.H >def.H
    !Filetag is 'out2' in this case
    type(axis)      , intent(in)           :: axisvar
    integer         , intent(in)           :: axisnr
    character(len=*), intent(in), optional :: filetag, axisname

    character(len=9)                       :: stringbank 
    character(len=1)                       :: axhandle 

    if(axisnr<1.or.axisnr>9) call seperr('readaxis: axis# must be from 1-9')

    stringbank = '123456789'
    axhandle   = stringbank(axisnr:axisnr)

    if(present(filetag)) then
       if(filetag=='in') then
          call seperr('Cannot write to input!')
       else if(filetag=='out') then
          write(0,*) 'writeaxis: Specify filetag only if is /= than "out"'
          call to_history('n'//axhandle, axisvar%n)
          call to_history('o'//axhandle, axisvar%o)
          call to_history('d'//axhandle, axisvar%d)
          if(present(axisname)) then
             call to_history('label'//axhandle, axisname)
          end if
       else
          call to_history('n'//axhandle, axisvar%n, filetag)
          call to_history('o'//axhandle, axisvar%o, filetag)
          call to_history('d'//axhandle, axisvar%d, filetag)
          if(present(axisname)) then
             call to_history('label'//axhandle, axisname, filetag)
          end if
       end if
    else
       call to_history('n'//axhandle, axisvar%n)
       call to_history('o'//axhandle, axisvar%o)
       call to_history('d'//axhandle, axisvar%d)
       if(present(axisname)) then
          call to_history('label'//axhandle, axisname)
       end if
    end if
    
  end subroutine writeaxis
!-----------------------------------------------------------------------
  function readcharpar(name,validvals,default)
    character(len=*)              , intent(in)::name
    character(len=*), dimension(:), intent(in)::validvals
    integer                       , intent(in)::default

    character(len=len(validvals))             ::readcharpar

    integer :: nv, i
    logical :: isfromlist
    
    nv = size(validvals,1)

    if(default<1.or.default>nv) &
    call seperr('pick the index of one of the elements in validvals')

    call from_param(name, readcharpar, validvals(default))

    isfromlist=.false.
    do i=1,nv
       if(readcharpar==validvals(i)) isfromlist=.true.
    end do
    if(.not.isfromlist) &
    call seperr('character parameter value not among the accepted choices')
  end function readcharpar
end module ooio
