module wacon_mod
! inoffensive (?) bug: should call clean_rfft instead of clean_fft
  use ooio, only: axis
  use fft_data_mod, only: fft_type, init_fft_1d, init_rfft_1d, fft_1d_data, rfft_1d_data, clean_fft
  implicit none

  private
  public :: wacon_init, wacon_mixdom, wacon_findif, clean_wacon
  public :: rft_fwd_axis1, rft_bck_axis1

  real,    parameter :: pi=3.1415926535898
  complex, parameter :: i = cmplx(0.,1.)
  integer            :: sign !sign of phaseshift in exponential
  integer            :: nx, nt
  real               :: dx, dz
  type(fft_type)     :: fft_1d_id, rft_1d_id
  type(axis)         :: w, k
  complex, dimension (:), allocatable :: C1,C2!high order zhang splitstep terms
  real,    dimension (:), allocatable :: lapco !Laplacian coefficients
  logical            :: amp_corr

  ! Bug danger! The common variables should be modified ONLY by init & clean. Or at worst, by a single subroutine only.

contains
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  function wacon_init(dowcon_as_shot, splitstep, zhang, x, z, t) result(omega)

    logical   , intent(in) :: dowcon_as_shot, splitstep, zhang
    type(axis), intent(in) :: x, z, t
    integer                :: iw, ik
    type(axis)             :: omega

    if(zhang) amp_corr=.true.
    sign=-1; if(dowcon_as_shot) sign = 1
    nx = x%n
    dx = x%d
    nt = t%n
    dz = z%d
    rft_1d_id = init_rfft_1d(nt)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    w%n = t%n/2
    w%d = 1./(t%n*t%d)
    w%o = w%d ! no zero frequency component
    allocate(w%all(w%n)); w%all=0.
    do iw = 1, w%n
       w%all(iw) = w%o  + (iw - 1) * w%d
    end do
    w%all = 2*pi * w%all

    omega = w
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    splitstep_or_not:if(splitstep) then
       fft_1d_id = init_fft_1d(nx)
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
       k%n = nx
       allocate(k%all(k%n)); k%all=0.

       k%d = 1./(k%n*dx)
       k%o = 1./(2*dx)-k%d !k does not increase uniformly!
       do ik=1, k%n
          k%all(ik) = mod(k%o+(ik-1)*k%d,1./dx)-k%o
       end do

       k%all = 2*pi * k%all

       if(amp_corr) then
          allocate (C1(nx), C2(nx)); C1=0; C2=0
       end if
    else
       allocate(lapco(3))
       lapco(1) = 0.5/(dx*dx)
       lapco(2) = -1./(dx*dx)
       lapco(3) = lapco(1)

    end if splitstep_or_not
  end function wacon_init
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  subroutine wacon_findif(C, v, boundcond, vz)
    ! Finite-difference downward continuation
    complex, dimension(:,:), intent(inout) :: C ! wavefield in (w,x) domain
    real   , dimension(nx) , intent(in)    :: v ! 0.5*interval velocity
    character(len=6),        intent(in)    :: boundcond
    real   , dimension(nx) , intent(in), optional :: vz ! velocity gradient
    ! tc = template coefficients. Holds the difference star for constant 
    ! velocity - first the three lower values, then the three upper ones
    complex, dimension(6, nx-2):: tc
    complex, dimension(   nx-2):: D
    real,    dimension(   nx  ):: n!=v/w
    integer                    :: iw
    real,    dimension(   nx-2):: E, F
    real,    parameter         :: alfa = 0.478242060 ! for 65 degrees (Yilmaz).
    ! alfa = 0.50 for 45 degrees and for 15 degrees equations
    real,    parameter         :: beta = 0.376369527 !for 65 degrees (Yilmaz). 
    ! beta = 0.25 for 45 degrees. beta = 0 for 15 degrees equation
    real,    parameter         :: sixth = 0.122996000000 ! 0 if we don't
    ! want the 1/6 trick. sixth = 1/6 to be literally a 1/6 trick. See IEI

    if(amp_corr.and..not.present(vz)) call seperr('vz needed for amp corr')
    if(.not.amp_corr.and.present(vz)) call seperr('vz NOT needed')

    freq:do iw = 1, w%n
       n = v / w%all(iw)
       E = 2.* n(2:nx-1)**2

       D = 2.* sixth * dx*dx + beta*E  -  i * sign * dz * alfa * n(2:nx-1)

       tc(1,:) = lapco(1) * D
       tc(2,:) = lapco(2) * D + 1.
       tc(3,:) = lapco(3) * D
       tc(4,:) = lapco(1) * conjg(D)
       tc(5,:) = lapco(2) * conjg(D) + 1.
       tc(6,:) = lapco(3) * conjg(D)

       call build_solve_trid_sys(C(iw,:), tc, boundcond)

       zhang_findif:if(amp_corr) then
          F = 0.25*vz*dz/v(2:nx-1)

          tc(1,:) = -lapco(1) * E
          tc(2,:) = -lapco(2) * E + 1. - F
          tc(3,:) = -lapco(3) * E
          tc(4,:) = -lapco(1) * E
          tc(5,:) = -lapco(2) * E + 1. + F
          tc(6,:) = -lapco(3) * E

          call build_solve_trid_sys(C(iw,:), tc, boundcond)
       end if zhang_findif

       C(iw,:) = C(iw,:) * exp(sign * i * dz / n)
    end do freq

  end subroutine wacon_findif
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  subroutine build_solve_trid_sys(c, tc, boundcond)
    complex, dimension(  nx  ),intent(inout):: c ! wavefield line at z and w
    complex, dimension(6,nx-2),intent(in)   :: tc ! template coefficients
    character(len=6),          intent(in)   :: boundcond

    complex, dimension(nx):: subdiag, middiag, sprdiag
    integer               :: ierr
    !need to be calculated every time because cgtsl modifies them

    subdiag(2:nx-1) = tc(1,:)
    middiag(2:nx-1) = tc(2,:)
    sprdiag(2:nx-1) = tc(3,:)

    subdiag(nx) = subdiag(nx-1)
    sprdiag( 1) = sprdiag(2)

    if (boundcond=='zslope') then
       middiag(1 ) = -sprdiag(1 )    
       middiag(nx) = -subdiag(nx)   
    else if (boundcond=='reflec') then
       middiag(1 ) = 10**30
       middiag(nx) = 10**30 
    end if

    ! c is recycled: we store the RHS in it:
    c(2:nx-1) = tc(4,:) * c(1:nx-2) + tc(5,:) * c(2:nx-1) + tc(6,:) * c(3:nx)
    c(1 ) = 0
    c(nx) = 0

    ! then we undo the trick - cgtsl transforms RHS into depth_row
    call cgtsl(nx,subdiag,middiag,sprdiag,c,ierr)
    if(ierr /= 0) call seperr("cgtsl: matrix needs pivoting ")

  end subroutine build_solve_trid_sys
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  subroutine wacon_mixdom(C, v, vz)
    ! splitstep downward continuation

    complex, dimension (:,:), intent(inout) :: C !wavefield in (w,x) domain
    real   , dimension (nx)  , intent(in)    :: v ! 0.5*interval velocity
    real   , dimension (nx)  , intent(in), optional :: vz ! velocity gradient

    real                                    :: v0! reference velocity
    real                                    :: sqa !sqrt(kz)
    integer                                 :: iw,ik
        
    if(amp_corr.and..not.present(vz)) call seperr('vz needed for amp corr')
    if(.not.amp_corr.and.present(vz)) call seperr('vz NOT needed')

    v0 = v(1)!=sum(v)/k%n

    freq:do iw = 1, w%n
       ! Maybe splitstep here? (L and G commutativity issue)
       call fft_1d_data(fft_1d_id, -1, k%n, C(iw,:))

       !Operations in the wavenumber domain
       splitstep_wvnmbr_dom:do ik = 1, k%n
          !---phase shift
          sqa = w%all(iw)**2 / v0**2 - k%all(ik)**2
          avoid_evan:if ( sqa > 0 ) then
             C(iw,ik) = C(iw,ik) * exp( i * sign * dz * sqrt(sqa) )
          else
             C(iw,ik) = 0.
          end if avoid_evan
       end do splitstep_wvnmbr_dom

       zhang3:if(amp_corr) then
             !---first order term
          C1 = C(iw,:) * ( k%all / w%all(iw) )**2
          !---second order term
          C2 = C(iw,:) * ( k%all / w%all(iw) )**4
       end if zhang3
       
       !Fourier transforms of all the orders of the wavefield
       call fft_1d_data(fft_1d_id, 1, k%n, C(iw,:))
       zhang2:if(amp_corr) then
          call fft_1d_data(fft_1d_id,1, k%n, C1)
          call fft_1d_data(fft_1d_id, 1, k%n, C2)
       end if zhang2
       
       !Operations in the space domain
       zhang1:if(amp_corr) then
          !---first order term
          C1 = C1 * ( -vz * v  * dz / 2. )
          !---second order term
          C2 = C2 * 0.5 * v*v * vz * dz * (vz*dz/4. - v)
          !---regroup all terms
          C(iw,:) = C(iw,:) + C1 + C2
          C(iw,:) = C(iw,:) * exp(-0.5 * vz * dz / v)
       end if zhang1

       !---split step
       ! Danger! Unknown whether this should be here or first thing at
       ! the beginning of the freq loop. Does Lambda commute with Gamma?
       C(iw,:) = C(iw,:) * exp(i*sign*dz*w%all(iw)*(1./v-1./v0))

    end do freq
  end subroutine wacon_mixdom
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  subroutine clean_wacon(splitstep)
    logical, intent(in):: splitstep

    call clean_fft(rft_1d_id)
    deallocate(w%all)
    if(splitstep) then
       call clean_fft(fft_1d_id)
       deallocate(k%all)
       if(amp_corr) deallocate(C1, C2)
    else
       deallocate(lapco)
    end if
  end subroutine clean_wacon
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  function rft_fwd_axis1(U)
    real   , dimension(:,:)        , intent(in) :: U
    complex, dimension(size(U,1)/2,  size(U,2)) :: rft_fwd_axis1
    complex, dimension(size(U,1)/2+1)           :: all_freq
    integer                                     :: ix

    do ix = 1, nx
       call rfft_1d_data(rft_1d_id, -1, nt, U(:,ix), all_freq)
       rft_fwd_axis1(:,ix) = all_freq(2:) ! Throw away the zero frequency
    end do
  end function rft_fwd_axis1
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  function rft_bck_axis1(C)
    complex, dimension(:,:), intent(in)        :: C
    real   , dimension(2*size(C,1), size(C,2)) :: rft_bck_axis1
    complex, dimension(1+size(C,1))            :: all_freq
    integer                                    :: ix

    all_freq (1)=0

    do ix = 1, nx
       all_freq(2:) = C(:,ix)
       call rfft_1d_data(rft_1d_id, 1, nt, rft_bck_axis1(:,ix), all_freq)
    end do
  end function rft_bck_axis1
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
end module wacon_mod
