program GenVzModel

  use sep, only: sep_init, from_param, to_history, sep_close, sep_write
  implicit none

  ! v(z) = v0+vz*(z-z0)
  real, parameter:: dz=4.572   ! 4.572 m   = 15 ft
  real, parameter:: v0=1524.   ! 1524 m/s  = 5000 ft/s
  real, parameter:: z0=763.524 ! 763.524 m = 2505 ft
  real, parameter:: vz=0.3     ! =dv/dz

  integer                          :: nx, nz, i
  real                             :: ox, dx, z, v
  real,  dimension(:), allocatable :: zslice

  call sep_init ()
  call from_param("nx", nx)
  call from_param("ox", ox)
  call from_param("dx", dx)
  call from_param("nz", nz)
  call to_history("n1", nx)
  call to_history("o1", ox)
  call to_history("d1", dx)
  call to_history("n2", nz)
  call to_history("o2", 0.)
  call to_history("d2", dz)
  call sep_close ()  

  allocate(zslice(nx)); zslice=1.

  do i=1,nz
     z = (i-1)*dz
     v = v0
     if (z>z0) then
        v = v + vz*(z-z0)
     end if
     call sep_write(v*zslice)
  end do
  deallocate(zslice)

end program GenVzModel
